"""Knowledge graph class."""

import asyncio
import logging
import os
from typing import Any, List, Optional

from dbgpt.core import Chunk, Embeddings, LLMClient
from dbgpt.core.awel.flow import Parameter, ResourceCategory, register_resource
from dbgpt.rag.transformer.keyword_extractor import KeywordExtractor
from dbgpt.storage.graph_store.base import GraphStoreBase, GraphStoreConfig
from dbgpt.storage.graph_store.graph import Graph
from dbgpt.storage.knowledge_graph.base import KnowledgeGraphBase
from dbgpt.storage.vector_store.filters import MetadataFilters
from dbgpt.util.i18n_utils import _
from dbgpt_ext.rag.transformer.triplet_extractor import TripletExtractor
from dbgpt_ext.storage.graph_store.factory import GraphStoreFactory
from dbgpt_ext.storage.graph_store.tugraph_store import TuGraphStoreConfig
from dbgpt_ext.storage.knowledge_graph.community.base import GraphStoreAdapter
from dbgpt_ext.storage.knowledge_graph.community.factory import GraphStoreAdapterFactory

logger = logging.getLogger(__name__)

GRAPH_PARAMETERS = [
    Parameter.build_from(
        _("Graph Name"),
        "name",
        str,
        description=_("The name of Graph, if not set, will use the default name."),
        optional=True,
        default="dbgpt_collection",
    ),
    Parameter.build_from(
        _("Embedding Function"),
        "embedding_fn",
        Embeddings,
        description=_(
            "The embedding function of vector store, if not set, will use "
            "the default embedding function."
        ),
        optional=True,
        default=None,
    ),
    Parameter.build_from(
        _("Max Chunks Once Load"),
        "max_chunks_once_load",
        int,
        description=_(
            "The max number of chunks to load at once. If your document is "
            "large, you can set this value to a larger number to speed up the loading "
            "process. Default is 10."
        ),
        optional=True,
        default=10,
    ),
    Parameter.build_from(
        _("Max Threads"),
        "max_threads",
        int,
        description=_(
            "The max number of threads to use. Default is 1. If you set "
            "this bigger than 1, please make sure your vector store is thread-safe."
        ),
        optional=True,
        default=1,
    ),
]


# @register_resource(
#     _("Builtin Graph Config"),
#     "knowledge_graph_config",
#     category=ResourceCategory.KNOWLEDGE_GRAPH,
#     description=_("knowledge graph config."),
#     parameters=[
#         *GRAPH_PARAMETERS,
#         Parameter.build_from(
#             _("Knowledge Graph Type"),
#             "graph_store_type",
#             str,
#             description=_("graph store type."),
#             optional=True,
#             default="TuGraph",
#         ),
#         Parameter.build_from(
#             _("LLM Client"),
#             "llm_client",
#             LLMClient,
#             description=_("llm client for extract graph triplets."),
#         ),
#         Parameter.build_from(
#             _("LLM Model Name"),
#             "model_name",
#             str,
#             description=_("llm model name."),
#             optional=True,
#             default=None,
#         ),
#     ],
# )
# @dataclass
# class BuiltinKnowledgeGraphConfig(KnowledgeGraphConfig):
#     """Builtin knowledge graph config."""
#
#     __type__ = "tugraph"
#
#     llm_model: Optional[str] = field(
#         default=None, metadata={"description": "llm model name."}
#     )
#
#     graph_type: Optional[str] = field(
#         default="TuGraph", metadata={"description": "graph store type."}
#     )


@register_resource(
    _("Builtin Knowledge Graph"),
    "builtin_knowledge_graph",
    category=ResourceCategory.KNOWLEDGE_GRAPH,
    description=_("Builtin Knowledge Graph."),
    parameters=[
        Parameter.build_from(
            _("Graph Store Config"),
            "config",
            GraphStoreConfig,
            description=_("graph store config."),
        ),
        Parameter.build_from(
            _("Graph Store Name"),
            "name",
            str,
            optional=True,
            default="dbgpt",
            description=_("Graph Store Name"),
        ),
        Parameter.build_from(
            _("LLM Client"),
            "llm_client",
            LLMClient,
            description=_("llm client for extract graph triplets."),
        ),
        Parameter.build_from(
            _("LLM Model Name"),
            "llm_model",
            str,
            description=_("kg extract llm model name."),
            optional=True,
            default=None,
        ),
    ],
)
class BuiltinKnowledgeGraph(KnowledgeGraphBase):
    """Builtin knowledge graph class."""

    def __init__(
        self,
        config: GraphStoreConfig = None,
        name: Optional[str] = "dbgpt",
        llm_client: Optional[LLMClient] = None,
        llm_model: Optional[str] = None,
    ):
        """Create builtin knowledge graph instance."""
        super().__init__()
        self._config = config
        self._llm_client = llm_client
        self._graph_name = name
        if not self._llm_client:
            raise ValueError("No llm client provided.")

        self._model_name = llm_model
        self._triplet_extractor = TripletExtractor(self._llm_client, self._model_name)
        self._keyword_extractor = KeywordExtractor(self._llm_client, self._model_name)
        self._graph_store: GraphStoreBase = self.__init_graph_store(config)
        self._graph_store_adapter: GraphStoreAdapter = self.__init_graph_store_adapter()

    def __init_graph_store(self, config: GraphStoreConfig) -> GraphStoreBase:
        def configure(cfg: GraphStoreConfig):
            cfg.name = self._graph_name

        graph_store_type = config.get_type_value() or os.getenv("GRAPH_STORE_TYPE")
        return GraphStoreFactory.create(graph_store_type, configure, config.to_dict())

    def __init_graph_store_adapter(self):
        return GraphStoreAdapterFactory.create(self._graph_store)

    def get_config(self) -> TuGraphStoreConfig:
        """Get the knowledge graph config."""
        return self._config

    @property
    def embeddings(self) -> Any:
        """Get the knowledge graph config."""
        return None

    def load_document(self, chunks: List[Chunk]) -> List[str]:
        """Extract and persist triplets to graph store."""

        async def process_chunk(chunk: Chunk):
            triplets = await self._triplet_extractor.extract(chunk.content)
            for triplet in triplets:
                self._graph_store_adapter.insert_triplet(*triplet)
            logger.info(f"load {len(triplets)} triplets from chunk {chunk.chunk_id}")
            return chunk.chunk_id

        # wait async tasks completed
        if not self.vector_name_exists():
            self._graph_store_adapter.create_graph(self._graph_name)
        tasks = [process_chunk(chunk) for chunk in chunks]
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        result = loop.run_until_complete(asyncio.gather(*tasks))
        loop.close()
        return result

    async def aload_document(
        self, chunks: List[Chunk], file_id: Optional[str] = None
    ) -> List[str]:  # type: ignore
        """Extract and persist triplets to graph store.

        Args:
            chunks: List[Chunk]: document chunks.
            file_id: Optional[str]: file id for document-level graph structure tracking.
        Return:
            List[str]: chunk ids.
        """
        if not self.vector_name_exists():
            self._graph_store_adapter.create_graph(self._graph_name)
        for chunk in chunks:
            triplets = await self._triplet_extractor.extract(chunk.content)
            for triplet in triplets:
                self._graph_store_adapter.insert_triplet(*triplet)
            logger.info(f"load {len(triplets)} triplets from chunk {chunk.chunk_id}")
        return [chunk.chunk_id for chunk in chunks]

    def similar_search_with_scores(
        self,
        text,
        topk,
        score_threshold: float,
        filters: Optional[MetadataFilters] = None,
    ) -> List[Chunk]:
        """Search neighbours on knowledge graph."""
        raise Exception("Sync similar_search_with_scores not supported")

    async def asimilar_search_with_scores(
        self,
        text,
        topk,
        score_threshold: float,
        filters: Optional[MetadataFilters] = None,
    ) -> List[Chunk]:
        """Search neighbours on knowledge graph."""
        if not filters:
            logger.info("Filters on knowledge graph not supported yet")

        # extract keywords and explore graph store
        keywords = await self._keyword_extractor.extract(text)
        subgraph = self._graph_store_adapter.explore_trigraph(
            keywords, limit=topk
        ).format()

        logger.info(f"Search subgraph from {len(keywords)} keywords")

        if not subgraph:
            return []

        content = (
            "The following entities and relationships provided after "
            "[Subgraph] are retrieved from the knowledge graph "
            "based on the keywords:\n"
            f'"{",".join(keywords)}".\n'
            "---------------------\n"
            "The following examples after [Entities] and [Relationships] that "
            "can help you understand the data format of the knowledge graph, "
            "but do not use them in the answer.\n"
            "[Entities]:\n"
            "(alice)\n"
            "(bob:{age:28})\n"
            '(carry:{age:18;role:"teacher"})\n\n'
            "[Relationships]:\n"
            "(alice)-[reward]->(alice)\n"
            '(alice)-[notify:{method:"email"}]->'
            '(carry:{age:18;role:"teacher"})\n'
            '(bob:{age:28})-[teach:{course:"math";hour:180}]->(alice)\n'
            "---------------------\n"
            f"[Subgraph]:\n{subgraph}\n"
        )
        return [Chunk(content=content)]

    def query_graph(self, limit: Optional[int] = None) -> Graph:
        """Query graph."""
        return self._graph_store_adapter.get_full_graph(limit)

    def truncate(self) -> List[str]:
        """Truncate knowledge graph."""
        logger.info(f"Truncate graph {self._graph_name}")
        self._graph_store_adapter.truncate()

        logger.info("Truncate keyword extractor")
        self._keyword_extractor.truncate()

        logger.info("Truncate triplet extractor")
        self._triplet_extractor.truncate()

        return [self._graph_name]

    def delete_vector_name(self, index_name: str):
        """Delete vector name."""
        logger.info(f"Drop graph {index_name}")
        self._graph_store_adapter.drop()

        logger.info("Drop keyword extractor")
        self._keyword_extractor.drop()

        logger.info("Drop triplet extractor")
        self._triplet_extractor.drop()

    def delete_by_ids(self, ids: str) -> List[str]:
        """Delete by ids."""
        self._graph_store_adapter.delete_document(chunk_id=ids)
        return []

    def vector_name_exists(self) -> bool:
        """Whether name exists."""
        return self._graph_store_adapter.graph_store.is_exist(self._graph_name)



"""Builtin Community metastore."""

import logging
from typing import List, Optional

from dbgpt.core import Chunk
from dbgpt.datasource.rdbms.base import RDBMSConnector
from dbgpt.storage.vector_store.base import VectorStoreBase
from dbgpt_ext.storage.knowledge_graph.community.base import (
    Community,
    CommunityMetastore,
)

logger = logging.getLogger(__name__)


class BuiltinCommunityMetastore(CommunityMetastore):
    """Builtin Community metastore."""

    def __init__(
        self,
        vector_store: VectorStoreBase,
        rdb_store: Optional[RDBMSConnector] = None,
        index_name: Optional[str] = None,
        max_chunks_once_load: Optional[int] = 10,
        max_threads: Optional[int] = 1,
        top_k: Optional[int] = 5,
        score_threshold: Optional[float] = 0.7,
    ):
        """Initialize Community metastore."""
        self._vector_store = vector_store
        self._rdb_store = rdb_store

        self._vector_space = index_name
        self._max_chunks_once_load = max_chunks_once_load
        self._max_threads = max_threads
        self._topk = top_k
        self._score_threshold = score_threshold

    def get(self, community_id: str) -> Community:
        """Get community."""
        raise NotImplementedError("Get community not allowed")

    def list(self) -> List[Community]:
        """Get all communities."""
        raise NotImplementedError("List communities not allowed")

    async def search(self, query: str) -> List[Community]:
        """Search communities relevant to query."""
        chunks = await self._vector_store.asimilar_search_with_scores(
            query, self._topk, self._score_threshold
        )
        return [Community(id=chunk.chunk_id, summary=chunk.content) for chunk in chunks]

    async def save(self, communities: List[Community]):
        """Save communities."""
        chunks = [
            Chunk(id=c.id, content=c.summary, metadata={"total": len(communities)})
            for c in communities
        ]
        await self._vector_store.aload_document_with_limit(
            chunks, self._max_chunks_once_load, self._max_threads
        )
        logger.info(f"Save {len(communities)} communities")

    async def truncate(self):
        """Truncate community metastore."""
        self._vector_store.truncate()

    def drop(self):
        """Drop community metastore."""
        if self._vector_store.vector_name_exists():
            self._vector_store.delete_vector_name(self._vector_space)



"""Vertica dialect."""

from __future__ import absolute_import, division, print_function

from .base import VerticaDialect as BaseVerticaDialect


# noinspection PyAbstractClass, PyClassHasNoInit
class VerticaDialect(BaseVerticaDialect):
    """Vertica dialect class."""

    driver = "vertica_python"
    # TODO: support SQL caching, for more info see:
    # https://docs.sqlalchemy.org/en/14/core/connections.html#caching-for-third-party-dialects
    supports_statement_cache = False
    # No lastrowid support. TODO support SELECT LAST_INSERT_ID();
    postfetch_lastrowid = False

    @classmethod
    def dbapi(cls):
        """Get Driver."""
        vertica_python = __import__("vertica_python")
        return vertica_python



"""CommunitySummarizer class."""

import logging

from dbgpt.core import LLMClient
from dbgpt.rag.transformer.llm_summarizer import LLMSummarizer

logger = logging.getLogger(__name__)


class CommunitySummarizer(LLMSummarizer):
    """CommunitySummarizer class."""

    def __init__(self, llm_client: LLMClient, model_name: str):
        """Initialize the CommunitySummaryExtractor."""
        super().__init__(llm_client, model_name, COMMUNITY_SUMMARY_PT_CN)


COMMUNITY_SUMMARY_PT_CN = (
    "## 角色\n"
    "你非常擅长知识图谱的信息总结，能根据给定的知识图谱中的实体和关系的名称以及描述"
    "信息，全面、恰当地对知识图谱子图信息做出总结性描述，并且不会丢失关键的信息。\n"
    "\n"
    "## 技能\n"
    "### 技能 1: 实体识别\n"
    "- 准确地识别[Entities:]章节中的实体信息，包括实体名、实体描述信息。\n"
    "- 实体信息的一般格式有:\n"
    "(实体名)\n"
    "(实体名:实体描述)\n"
    "(实体名:实体属性表)\n"
    "(文本块ID:文档块内容)\n"
    "(目录ID:目录名)\n"
    "(文档ID:文档名称)\n"
    "\n"
    "### 技能 2: 关系识别\n"
    "- 准确地识别[Relationships:]章节中的关系信息，包括来源实体名、关系名、"
    "目标实体名、关系描述信息，实体名也可能是文档ID、目录ID、文本块ID。\n"
    "- 关系信息的一般格式有:\n"
    "(来源实体名)-[关系名]->(目标实体名)\n"
    "(来源实体名)-[关系名:关系描述]->(目标实体名)\n"
    "(来源实体名)-[关系名:关系属性表]->(目标实体名)\n"
    "(文本块ID)-[包含]->(实体名)\n"
    "(目录ID)-[包含]->(文本块实体)\n"
    "(目录ID)-[包含]->(子目录ID)\n"
    "(文档ID)-[包含]->(文本块实体)\n"
    "(文档ID)-[包含]->(目录ID)\n"
    "\n"
    "### 技能 3: 图结构理解\n"
    "--请按照如下步骤理解图结构--\n"
    "1. 正确地将关系信息中的来源实体名与实体信息关联。\n"
    "2. 正确地将关系信息中的目标实体名与实体信息关联。\n"
    "3. 根据提供的关系信息还原出图结构。"
    "\n"
    "### 技能 4: 知识图谱总结\n"
    "--请按照如下步骤总结知识图谱--\n"
    "1. 确定知识图谱表达的主题或话题，突出关键实体和关系。"
    "2. 使用准确、恰当、简洁的语言总结图结构表达的信息，不要生成与图结构中无关的信息。"
    "\n"
    "## 约束条件\n"
    "- 不要在答案中描述你的思考过程，直接给出用户问题的答案，不要生成无关信息。\n"
    "- 确保以第三人称书写，从客观角度对知识图谱表达的信息进行总结性描述。\n"
    "- 如果实体或关系的描述信息为空，对最终的总结信息没有贡献，不要生成无关信息。\n"
    "- 如果提供的描述信息相互矛盾，请解决矛盾并提供一个单一、连贯的描述。\n"
    "- 避免使用停用词和过于常见的词汇。\n"
    "\n"
    "## 参考案例\n"
    "--案例仅帮助你理解提示词的输入和输出格式，请不要在答案中使用它们。--\n"
    "输入:\n"
    "```\n"
    "Entities:\n"
    "(菲尔・贾伯#菲尔兹咖啡创始人)\n"
    "(菲尔兹咖啡#加利福尼亚州伯克利创立的咖啡品牌)\n"
    "(雅各布・贾伯#菲尔・贾伯的儿子)\n"
    "(美国多地#菲尔兹咖啡的扩展地区)\n"
    "\n"
    "Relationships:\n"
    "(菲尔・贾伯#创建#菲尔兹咖啡#1978年在加利福尼亚州伯克利创立)\n"
    "(菲尔兹咖啡#位于#加利福尼亚州伯克利#菲尔兹咖啡的创立地点)\n"
    "(菲尔・贾伯#拥有#雅各布・贾伯#菲尔・贾伯的儿子)\n"
    "(雅各布・贾伯#担任#首席执行官#在2005年成为菲尔兹咖啡的首席执行官)\n"
    "(菲尔兹咖啡#扩展至#美国多地#菲尔兹咖啡的扩展范围)\n"
    "```\n"
    "\n"
    "输出:\n"
    "```\n"
    "菲尔兹咖啡是由菲尔・贾伯在1978年于加利福尼亚州伯克利创立的咖啡品牌。"
    "菲尔・贾伯的儿子雅各布・贾伯在2005年接任首席执行官，领导公司扩展到了美国多地，"
    "进一步巩固了菲尔兹咖啡作为加利福尼亚州伯克利创立的咖啡品牌的市场地位。\n"
    "```\n"
    "\n"
    "----\n"
    "\n"
    "请根据接下来[知识图谱]提供的信息，按照上述要求，总结知识图谱表达的信息。\n"
    "\n"
    "[知识图谱]:\n"
    "{graph}\n"
    "\n"
    "[总结]:\n"
    "\n"
)

COMMUNITY_SUMMARY_PT_EN = (
    "## Role\n"
    "You are highly skilled in summarizing information from knowledge graphs. "
    "Based on the names and descriptions of entities and relationships in a "
    "given knowledge graph, you can comprehensively and appropriately summarize"
    " the information of the subgraph without losing critical details.\n"
    "\n"
    "## Skills\n"
    "### Skill 1: Entity Recognition\n"
    "- Accurately recognize entity information in the [Entities:] section, "
    "including entity names and descriptions.\n"
    "- The general formats for entity information are:\n"
    "(entity_name)\n"
    "(entity_name: entity_description)\n"
    "(entity_name: entity_property_map)\n"
    "(chunk_id: chunk_content)\n"
    "(catalog_id: catalog_name)\n"
    "(document_id: document_name)\n"
    "\n"
    "### Skill 2: Relationship Recognition\n"
    "- Accurately recognize relationship information in the [Relationships:] "
    "section, including source_entity_name, relationship_name, "
    "target_entity_name, and relationship_description, The entity_name may "
    "also be the document_id, catalog_id, or chunk_id.\n"
    "- The general formats for relationship information are:\n"
    "(source_entity_name)-[relationship_name]->(target_entity_name)\n"
    "(source_entity_name)-[relationship_name: relationship_description]->"
    "(target_entity_name)\n"
    "(source_entity_name)-[relationship_name: relationship_property_map]->"
    "(target_entity_name)\n"
    "(chunk_id)-[Contains]->(entity_name)\n"
    "(catalog_id)-[Contains]->(chunk_id)\n"
    "(catalog_id)-[Contains]->(sub_catalog_id)\n"
    "(document_id)-[Contains]->(chunk_id)\n"
    "(document_id)-[Contains]->(catalog_id)\n"
    "\n"
    "### Skill 3: Graph Structure Understanding\n"
    "--Follow these steps to understand the graph structure--\n"
    "1. Correctly associate the source entity name in the "
    "relationship information with the entity information.\n"
    "2. Correctly associate the target entity name in the "
    "relationship information with the entity information.\n"
    "3. Reconstruct the graph structure based on the provided "
    "relationship information."
    "\n"
    "### Skill 4: Knowledge Graph Summarization\n"
    "--Follow these steps to summarize the knowledge graph--\n"
    "1. Determine the theme or topic expressed by the knowledge graph, "
    "highlighting key entities and relationships."
    "2. Use accurate, appropriate, and concise language to summarize "
    "the information expressed by the graph "
    "without generating irrelevant information."
    "\n"
    "## Constraints\n"
    "- Don't describe your thought process in the answer, provide the answer "
    "to the user's question directly without generating irrelevant information."
    "- Ensure the summary is written in the third person and objectively "
    "reflects the information conveyed by the knowledge graph.\n"
    "- If the descriptions of entities or relationships are empty and "
    "contribute nothing to the final summary, "
    "do not generate unrelated information.\n"
    "- If the provided descriptions are contradictory, resolve the conflicts "
    "and provide a single, coherent description.\n"
    "- Avoid using stop words and overly common words.\n"
    "\n"
    "## Reference Example\n"
    "--The case is only to help you understand the input and output format of "
    "the prompt, please do not use it in your answer.--\n"
    "Input:\n"
    "```\n"
    "Entities:\n"
    "(Phil Jaber#Founder of Philz Coffee)\n"
    "(Philz Coffee#Coffee brand founded in Berkeley, California)\n"
    "(Jacob Jaber#Son of Phil Jaber)\n"
    "(Multiple locations in the USA#Expansion regions of Philz Coffee)\n"
    "\n"
    "Relationships:\n"
    "(Phil Jaber#Created#Philz Coffee"
    "#Founded in Berkeley, California in 1978)\n"
    "(Philz Coffee#Located in#Berkeley, California"
    "#Founding location of Philz Coffee)\n"
    "(Phil Jaber#Has#Jacob Jaber#Son of Phil Jaber)\n"
    "(Jacob Jaber#Serves as#CEO#Became CEO of Philz Coffee in 2005)\n"
    "(Philz Coffee#Expanded to#Multiple locations in the USA"
    "#Expansion regions of Philz Coffee)\n"
    "```\n"
    "\n"
    "Output:\n"
    "```\n"
    "Philz Coffee is a coffee brand founded by Phil Jaber in 1978 in "
    "Berkeley, California. Phil Jaber's son, Jacob Jaber, took over as CEO in "
    "2005, leading the company to expand to multiple locations in the USA, "
    "further solidifying Philz Coffee's market position as a coffee brand "
    "founded in Berkeley, California.\n"
    "```\n"
    "\n"
    "----\n"
    "\n"
    "Please summarize the information expressed by the [KnowledgeGraph] "
    "provided in the following section according to the above requirements.\n"
    "\n"
    "[KnowledgeGraph]:\n"
    "{graph}\n"
    "\n"
    "[Summary]:\n"
    "\n"
)



from dataclasses import dataclass, field
from typing import Any, Dict, Optional

from dbgpt.core.interface.file import StorageBackend, StorageBackendConfig
from dbgpt.util.i18n_utils import _


@dataclass
class S3StorageConfig(StorageBackendConfig):
    __type__ = "s3"
    endpoint: str = field(
        metadata={
            "help": _(
                "The endpoint of the s3 server. e.g. https://s3.us-east-1.amazonaws.com"
            )
        },
    )
    region: str = field(
        metadata={"help": _("The region of the s3 server. e.g. us-east-1")},
    )
    access_key_id: Optional[str] = field(
        default=None,
        metadata={
            "help": _(
                "The access key ID of the s3 server. You can also set it in the "
                "environment variable AWS_ACCESS_KEY_ID"
            ),
            "tags": "privacy",
        },
    )
    access_key_secret: Optional[str] = field(
        default=None,
        metadata={
            "help": _(
                "The access key secret of the s3 server. You can also set it in the "
                "environment variable AWS_SECRET_ACCESS_KEY"
            ),
            "tags": "privacy",
        },
    )
    use_environment_credentials: Optional[bool] = field(
        default=False,
        metadata={
            "help": _(
                "Whether to use the environment variables AWS_ACCESS_KEY_ID and "
                "AWS_SECRET_ACCESS_KEY as the credentials. Default is False."
            ),
        },
    )
    fixed_bucket: Optional[str] = field(
        default=None,
        metadata={
            "help": _(
                "The fixed bucket name to use. If set, all logical buckets in DB-GPT "
                "will be mapped to this bucket. We suggest you set this value to avoid "
                "bucket name conflicts."
            )
        },
    )
    bucket_prefix: Optional[str] = field(
        default="dbgpt-fs-",
        metadata={
            "help": _(
                "The prefix of the bucket name. If set, all logical buckets in DB-GPT "
                "will be prefixed with this value. Just work when fixed_bucket is None."
            )
        },
    )
    auto_create_bucket: Optional[bool] = field(
        default=True,
        metadata={
            "help": _(
                "Whether to create the bucket automatically if it does not exist. "
                "If set to False, the bucket must exist before using it."
            )
        },
    )
    save_chunk_size: Optional[int] = field(
        default=1024 * 1024,
        metadata={
            "help": _(
                "The chunk size when saving the file. When the file is larger 10x than "
                "this value, it will be uploaded in multiple parts. Default is 1M."
            )
        },
    )
    signature_version: Optional[str] = field(
        default=None,
        metadata={
            "help": _(
                "The signature version of the s3 server. "
                "e.g. s3v4, s3v2, None (default)"
            )
        },
    )
    s3_config: Optional[Dict[str, Any]] = field(
        default_factory=dict,
        metadata={
            "help": _("The additional configuration for the S3 client."),
        },
    )

    def create_storage(self) -> StorageBackend:
        from .s3_storage import S3Storage

        return S3Storage(
            endpoint_url=self.endpoint,
            region_name=self.region,
            access_key_id=self.access_key_id,
            secret_access_key=self.access_key_secret,
            use_environment_credentials=self.use_environment_credentials,
            fixed_bucket=self.fixed_bucket,
            bucket_prefix=self.bucket_prefix,
            auto_create_bucket=self.auto_create_bucket,
            save_chunk_size=self.save_chunk_size,
            signature_version=self.signature_version,
            s3_config=self.s3_config,
        )



"""Neo4j store."""

import logging
import os
from dataclasses import dataclass, field

from dbgpt.core.awel.flow import Parameter, ResourceCategory, register_resource
from dbgpt.storage.graph_store.base import GraphStoreBase, GraphStoreConfig
from dbgpt.storage.graph_store.graph import GraphElemType
from dbgpt.util.i18n_utils import _
from dbgpt_ext.datasource.conn_neo4j import Neo4jConnector

logger = logging.getLogger(__name__)


@register_resource(
    _("Neo4j Graph Config"),
    "neo4j_config",
    category=ResourceCategory.KNOWLEDGE_GRAPH,
    description=_("Neo4j config."),
    parameters=[
        Parameter.build_from(
            _("host"),
            "host",
            str,
            optional=True,
            default="127.0.0.1",
            description=_("Neo4j host"),
        ),
        Parameter.build_from(
            _("port"),
            "port",
            int,
            optional=True,
            default=7687,
            description=_("Neo4j port"),
        ),
        Parameter.build_from(
            _("username"),
            "username",
            str,
            optional=True,
            default="neo4j",
            description=_("Neo4j username"),
        ),
        Parameter.build_from(
            _("password"),
            "password",
            str,
            optional=True,
            default="neo4j",
            description=_("Neo4j password"),
        ),
        Parameter.build_from(
            _("database"),
            "database",
            str,
            optional=True,
            default="neo4j",
            description=_("Neo4j database name"),
        ),
    ],
)
@dataclass
class Neo4jStoreConfig(GraphStoreConfig):
    """Neo4j store config."""

    __type__ = "neo4j"

    host: str = field(
        default="127.0.0.1",
        metadata={
            "description": "Neo4j host",
        },
    )
    port: int = field(
        default=7687,
        metadata={
            "description": "Neo4j port",
        },
    )
    username: str = field(
        default="neo4j",
        metadata={
            "description": "login username",
        },
    )
    password: str = field(
        default="neo4j",
        metadata={
            "description": "login password",
        },
    )
    database: str = field(
        default="neo4j",
        metadata={
            "description": "Neo4j database name",
        },
    )
    vertex_type: str = field(
        default=GraphElemType.ENTITY.value,
        metadata={
            "description": "The type of vertex, `entity` by default.",
        },
    )
    document_type: str = field(
        default=GraphElemType.DOCUMENT.value,
        metadata={
            "description": "The type of document vertex, `document` by default.",
        },
    )
    chunk_type: str = field(
        default=GraphElemType.CHUNK.value,
        metadata={
            "description": "The type of chunk vertex, `chunk` by default.",
        },
    )
    edge_type: str = field(
        default=GraphElemType.RELATION.value,
        metadata={
            "description": "The type of relation edge, `relation` by default.",
        },
    )
    include_type: str = field(
        default=GraphElemType.INCLUDE.value,
        metadata={
            "description": "The type of include edge, `include` by default.",
        },
    )
    next_type: str = field(
        default=GraphElemType.NEXT.value,
        metadata={
            "description": "The type of next edge, `next` by default.",
        },
    )
    enable_summary: bool = field(
        default=True,
        metadata={
            "description": "Enable graph community summary or not.",
        },
    )
    enable_similarity_search: bool = field(
        default=False,
        metadata={
            "description": "Enable the similarity search or not",
        },
    )


class Neo4jStore(GraphStoreBase):
    """Neo4j graph store."""

    def __init__(self, config: Neo4jStoreConfig) -> None:
        """Initialize the Neo4jStore with connection details."""
        self._config = config
        self._host = config.host or os.getenv("NEO4J_HOST", "127.0.0.1")
        self._port = int(config.port or os.getenv("NEO4J_PORT", "7687"))
        self._username = config.username or os.getenv("NEO4J_USER", "neo4j")
        self._password = config.password or os.getenv("NEO4J_PASSWORD", "neo4j")
        self._database = config.database or os.getenv("NEO4J_DATABASE", "neo4j")
        self.enable_summary = config.enable_summary or (
            os.getenv("GRAPH_COMMUNITY_SUMMARY_ENABLED", "").lower() == "true"
        )
        self.enable_similarity_search = config.enable_similarity_search or (
            os.getenv("SIMILARITY_SEARCH_ENABLED", "").lower() == "true"
        )

        self._graph_name = config.name

        self.conn = Neo4jConnector.from_uri_db(
            host=self._host,
            port=self._port,
            user=self._username,
            pwd=self._password,
            db_name=self._database,  # Use database field instead of config.name
        )

    def get_config(self) -> Neo4jStoreConfig:
        """Get the Neo4j store config."""
        return self._config

    def is_exist(self, name: str) -> bool:
        """Check if graph (database) exists in Neo4j.

        Args:
            name: Database name to check

        Returns:
            bool: True if database exists
        """
        return self.conn.is_exist(name)



from unittest.mock import MagicMock

import pytest

from dbgpt.core import Chunk
from dbgpt.rag.retriever.embedding import EmbeddingRetriever


@pytest.fixture
def top_k():
    return 4


@pytest.fixture
def query():
    return "test query"


@pytest.fixture
def mock_vector_store_connector():
    return MagicMock()


@pytest.fixture
def embedding_retriever(top_k, mock_vector_store_connector):
    return EmbeddingRetriever(
        top_k=top_k,
        query_rewrite=None,
        index_store=mock_vector_store_connector,
    )


def test_retrieve(query, top_k, mock_vector_store_connector, embedding_retriever):
    expected_chunks = [Chunk() for _ in range(top_k)]
    mock_vector_store_connector.similar_search.return_value = expected_chunks

    retrieved_chunks = embedding_retriever._retrieve(query)

    assert len(retrieved_chunks) == top_k



from unittest.mock import mock_open, patch

import pytest

from ..txt import TXTKnowledge

MOCK_TXT_CONTENT = b"Sample text content for testing.\nAnother line of text."

MOCK_CHARDET_RESULT = {"encoding": "utf-8", "confidence": 0.99}


@pytest.fixture
def mock_file_open():
    with patch(
        "builtins.open", mock_open(read_data=MOCK_TXT_CONTENT), create=True
    ) as mock_file:
        yield mock_file


@pytest.fixture
def mock_chardet_detect():
    with patch("chardet.detect", return_value=MOCK_CHARDET_RESULT) as mock_detect:
        yield mock_detect


# 定义测试函数
def test_load_from_txt(mock_file_open, mock_chardet_detect):
    file_path = "test_document.txt"
    knowledge = TXTKnowledge(file_path=file_path)
    documents = knowledge._load()

    assert len(documents) == 1
    assert "Sample text content for testing." in documents[0].content
    assert documents[0].metadata["source"] == file_path

    mock_file_open.assert_called_once_with(file_path, "rb")

    mock_chardet_detect.assert_called_once()



"""Graph Store Module."""

from typing import Tuple, Type


def _import_memgraph() -> Tuple[Type, Type]:
    from dbgpt.storage.graph_store.memgraph_store import (
        MemoryGraphStore,
        MemoryGraphStoreConfig,
    )

    return MemoryGraphStore, MemoryGraphStoreConfig


def _import_tugraph() -> Tuple[Type, Type]:
    from dbgpt_ext.storage.graph_store.tugraph_store import (
        TuGraphStore,
        TuGraphStoreConfig,
    )

    return TuGraphStore, TuGraphStoreConfig


def _import_neo4j() -> Tuple[Type, Type]:
    from dbgpt_ext.storage.graph_store.neo4j_store import Neo4jStore, Neo4jStoreConfig

    return Neo4jStore, Neo4jStoreConfig


def __getattr__(name: str) -> Tuple[Type, Type]:
    if name == "Memory":
        return _import_memgraph()
    elif name == "TuGraph":
        return _import_tugraph()
    elif name == "Neo4j":
        return _import_neo4j()
    else:
        raise AttributeError(f"Could not find: {name}")


__all__ = ["Memory", "TuGraph", "Neo4j"]



from unittest.mock import mock_open, patch

import pytest

from ..html import HTMLKnowledge

MOCK_HTML_CONTENT = b"""
<html>
<head>
<title>Test HTML</title>
</head>
<body>
<p>This is a paragraph.</p>
</body>
</html>
"""

MOCK_CHARDET_RESULT = {"encoding": "utf-8", "confidence": 0.99}


@pytest.fixture
def mock_file_open():
    with patch(
        "builtins.open", mock_open(read_data=MOCK_HTML_CONTENT), create=True
    ) as mock_file:
        yield mock_file


@pytest.fixture
def mock_chardet_detect():
    with patch("chardet.detect", return_value=MOCK_CHARDET_RESULT) as mock_detect:
        yield mock_detect


def test_load_from_html(mock_file_open, mock_chardet_detect):
    file_path = "test_document.html"
    knowledge = HTMLKnowledge(file_path=file_path)
    documents = knowledge._load()

    assert len(documents) == 1
    assert "This is a paragraph." in documents[0].content
    assert documents[0].metadata["source"] == file_path

    mock_file_open.assert_called_once_with(file_path, "rb")

    mock_chardet_detect.assert_called_once()



from unittest.mock import MagicMock, mock_open, patch

import pytest

from ..pdf import PDFKnowledge

MOCK_PDF_PAGES = [
    ("", 0),
    ("", 1),
]


@pytest.fixture
def mock_pdf_open_and_reader():
    mock_pdf_file = mock_open()
    mock_reader = MagicMock()
    mock_reader.pages = [
        MagicMock(extract_text=MagicMock(return_value=page[0]))
        for page in MOCK_PDF_PAGES
    ]
    with patch("builtins.open", mock_pdf_file):
        with patch("pdfplumber.open", return_value=mock_reader) as mock:
            yield mock


def test_load_from_pdf(mock_pdf_open_and_reader):
    file_path = "test_document"
    knowledge = PDFKnowledge(file_path=file_path)
    documents = knowledge._load()

    assert len(documents) == len(MOCK_PDF_PAGES)
    for i, document in enumerate(documents):
        assert MOCK_PDF_PAGES[i][0] in document.content
        assert document.metadata["title"] == file_path
        assert document.metadata["type"] == "text"

    #



"""Ocean base connect."""

import logging
from dataclasses import dataclass, field
from typing import Type

from dbgpt.core.awel.flow import (
    TAGS_ORDER_HIGH,
    ResourceCategory,
    auto_register_resource,
)
from dbgpt.datasource.rdbms.base import RDBMSConnector, RDBMSDatasourceParameters
from dbgpt.util.i18n_utils import _

logger = logging.getLogger(__name__)


@auto_register_resource(
    label=_("OceanBase datasource"),
    category=ResourceCategory.DATABASE,
    tags={"order": TAGS_ORDER_HIGH},
    description=_("An Ultra-Fast & Cost-Effective Distributed SQL Database."),
)
@dataclass
class OceanBaseParameters(RDBMSDatasourceParameters):
    """Oceanbase connection parameters."""

    __type__ = "oceanbase"
    driver: str = field(
        default="mysql+ob",
        metadata={
            "help": _("Driver name for oceanbase, default is mysql+ob."),
        },
    )

    def create_connector(self) -> "OceanBaseConnector":
        return OceanBaseConnector.from_parameters(self)


class OceanBaseConnector(RDBMSConnector):
    """Connect Oceanbase Database fetch MetaData.

    Args:
    Usage:
    """

    db_type: str = "oceanbase"
    db_dialect: str = "mysql"
    driver: str = "mysql+ob"

    default_db = ["information_schema", "performance_schema", "sys", "mysql"]

    @classmethod
    def param_class(cls) -> Type[OceanBaseParameters]:
        """Return the parameter class."""
        return OceanBaseParameters

    def get_users(self):
        """Get_users."""
        return []

    def get_grants(self):
        """Get_grants."""
        return []

    def get_collation(self):
        """Get collation."""
        return "UTF-8"

    def get_charset(self):
        """Get_charset."""
        return "UTF-8"



"""Pre text splitter."""

from typing import Iterable, List

from dbgpt.core import Chunk, Document
from dbgpt.rag.text_splitter.text_splitter import TextSplitter


def _single_document_split(
    document: Document, pre_separator: str
) -> Iterable[Document]:
    origin_content = document.content
    for i, content in enumerate(origin_content.split(pre_separator)):
        metadata = document.metadata.copy()
        if "source" in metadata:
            metadata["source"] = metadata["source"] + "_pre_split_" + str(i)
        yield Chunk(content=content, metadata=metadata)


class PreTextSplitter(TextSplitter):
    """Split text by pre separator."""

    def __init__(self, pre_separator: str, text_splitter_impl: TextSplitter):
        """Create the pre text splitter instance.

        Args:
            pre_separator: pre separator
            text_splitter_impl: text splitter impl
        """
        self.pre_separator = pre_separator
        self._impl = text_splitter_impl

    def split_text(self, text: str, **kwargs) -> List[str]:
        """Split text by pre separator."""
        return self._impl.split_text(text)

    def split_documents(self, documents: Iterable[Document], **kwargs) -> List[Chunk]:
        """Split documents by pre separator."""

        def generator() -> Iterable[Document]:
            for doc in documents:
                yield from _single_document_split(doc, pre_separator=self.pre_separator)

        return self._impl.split_documents(generator())



"""Tree-based document retriever."""

import logging
from concurrent.futures import Executor, ThreadPoolExecutor
from typing import List, Optional

from dbgpt.core import Document
from dbgpt.rag.retriever import BaseRetriever, DefaultRanker, QueryRewrite, Ranker
from dbgpt.rag.transformer.base import ExtractorBase
from dbgpt.storage.vector_store.filters import MetadataFilters

logger = logging.getLogger(__name__)

RETRIEVER_NAME = "doc_tree_retriever"
TITLE = "title"
HEADER1 = "Header1"
HEADER2 = "Header2"
HEADER3 = "Header3"
HEADER4 = "Header4"
HEADER5 = "Header5"
HEADER6 = "Header6"


class TreeNode:
    """TreeNode class to represent a node in the document tree."""

    def __init__(
        self, node_id: str, level_text: str, level: int, content: Optional[str] = None
    ):
        """Initialize a TreeNode."""
        self.node_id = node_id
        self.level = level  # 0: title, 1: header1, 2: header2, 3: header3
        self.level_text = level_text  # 0: title, 1: header1, 2: header2, 3: header3
        self.children = []
        self.content = content
        self.retriever = RETRIEVER_NAME

    def add_child(self, child_node):
        """Add a child node to the current node."""
        self.children.append(child_node)


class DocTreeIndex:
    def __init__(self):
        """Initialize the document tree index."""
        self.root = TreeNode("root_id", "Root", -1)

    def add_nodes(
        self,
        node_id: str,
        title: str,
        header1: Optional[str] = None,
        header2: Optional[str] = None,
        header3: Optional[str] = None,
        header4: Optional[str] = None,
        header5: Optional[str] = None,
        header6: Optional[str] = None,
        content: Optional[str] = None,
    ):
        """Add nodes to the document tree.

        Args:
            node_id (str): The ID of the node.
            title (str): The title of the document.
            header1 (Optional[str]): The first header.
            header2 (Optional[str]): The second header.
            header3 (Optional[str]): The third header.
            header4 (Optional[str]): The fourth header.
            header5 (Optional[str]): The fifth header.
            header6 (Optional[str]): The sixth header.
            content (Optional[str]): The content of the node.
        """
        # Assuming titles is a dictionary containing title and headers
        title_node = None
        if title:
            title_nodes = self.get_node_by_level(0)
            if not title_nodes:
                # If title already exists, do not add it again
                title_node = TreeNode(node_id, title, 0, content)
                self.root.add_child(title_node)
            else:
                title_node = title_nodes[0]
        current_node = title_node
        headers = [header1, header2, header3, header4, header5, header6]
        for level, header in enumerate(headers, start=1):
            if header:
                header_nodes = self.get_node_by_level_text(header)
                if header_nodes:
                    # If header already exists, do not add it again
                    current_node = header_nodes[0]
                    continue
                new_header_node = TreeNode(node_id, header, level, content)
                current_node.add_child(new_header_node)
                current_node = new_header_node

    def add_nodes_with_content(
        self,
        node_id: str,
        title: str,
        header1: Optional[str] = None,
        header2: Optional[str] = None,
        header3: Optional[str] = None,
        header4: Optional[str] = None,
        header5: Optional[str] = None,
        header6: Optional[str] = None,
    ):
        """Add nodes to the document tree.

        Args:
            node_id (str): The ID of the node.
            title (str): The title of the document.
            header1 (Optional[str]): The first header.
            header2 (Optional[str]): The second header.
            header3 (Optional[str]): The third header.
            header4 (Optional[str]): The fourth header.
            header5 (Optional[str]): The fifth header.
            header6 (Optional[str]): The sixth header.
        """
        # Assuming titles is a dictionary containing title and headers
        title_node = None
        if title:
            title_nodes = self.get_node_by_level(0)
            if not title_nodes:
                # If title already exists, do not add it again
                title_node = TreeNode(node_id, title, 0)
                self.root.add_child(title_node)
            else:
                title_node = title_nodes[0]
        current_node = title_node
        headers = [header1, header2, header3, header4, header5, header6]
        for level, header in enumerate(headers, start=1):
            if header:
                header_nodes = self.get_node_by_level_text(header)
                if header_nodes:
                    # If header already exists, do not add it again
                    current_node = header_nodes[0]
                    continue
                new_header_node = TreeNode(node_id, header, level)
                current_node.add_child(new_header_node)
                current_node = new_header_node

    def get_node_by_level(self, level):
        """Get nodes by level."""
        # Traverse the tree to find nodes at the specified level
        result = []
        self._traverse(self.root, level, result)
        return result

    def get_node_by_level_text(self, content):
        """Get nodes by level."""
        # Traverse the tree to find nodes at the specified level
        result = []
        self._traverse_by_level_text(self.root, content, result)
        return result

    def get_all_children(self, node):
        """get all children of the node."""
        # Get all children of the current node
        result = []
        self._traverse(node, node.level, result)
        return result

    def display_tree(self, node: TreeNode, prefix: Optional[str] = ""):
        """Recursive function to display the directory structure with visual cues."""
        # Print the current node title with prefix
        if node.content:
            print(
                f"{prefix}├── {node.level_text} (node_id: {node.node_id}) "
                f"(content: {node.content})"
            )
        else:
            print(f"{prefix}├── {node.level_text} (node_id: {node.node_id})")

        # Update prefix for children
        new_prefix = prefix + "│   "  # Extend the prefix for child nodes
        for i, child in enumerate(node.children):
            if i == len(node.children) - 1:  # If it's the last child
                new_prefix_child = prefix + "└── "
            else:
                new_prefix_child = new_prefix

            # Recursive call for the next child node
            self.display_tree(child, new_prefix_child)

    def _traverse(self, node, level, result):
        """Traverse the tree to find nodes at the specified level."""
        # If the current node's level matches the specified level, add it to the result
        if node.level == level:
            result.append(node)
        # Recursively traverse child nodes
        for child in node.children:
            self._traverse(child, level, result)

    def _traverse_by_level_text(self, node, level_text, result):
        """Traverse the tree to find nodes at the specified level."""
        # If the current node's level matches the specified level, add it to the result
        if node.level_text == level_text:
            result.append(node)
        # Recursively traverse child nodes
        for child in node.children:
            self._traverse_by_level_text(child, level_text, result)

    def search_keywords(self, node, keyword) -> Optional[TreeNode]:
        # Check if the keyword matches the current node title
        if keyword.lower() == node.level_text.lower():
            logger.info(f"DocTreeIndex Match found in: {node.level_text}")
            return node
        # Recursively search in child nodes
        for child in node.children:
            result = self.search_keywords(child, keyword)
            if result:
                return result
        # Check if the keyword matches any of the child nodes
        # If no match, continue to search in all children
        return None


class DocTreeRetriever(BaseRetriever):
    """Doc Tree retriever."""

    def __init__(
        self,
        docs: List[Document] = None,
        top_k: Optional[int] = 10,
        query_rewrite: Optional[QueryRewrite] = None,
        rerank: Optional[Ranker] = None,
        keywords_extractor: Optional[ExtractorBase] = None,
        with_content: bool = False,
        show_tree: bool = True,
        executor: Optional[Executor] = None,
    ):
        """Create DocTreeRetriever.

        Args:
            docs (List[Document]): List of documents to initialize the tree with.
            top_k (int): top k
            query_rewrite (Optional[QueryRewrite]): query rewrite
            rerank (Ranker): rerank
            keywords_extractor (Optional[ExtractorBase]): keywords extractor
            with_content: bool: whether to include content
            executor (Optional[Executor]): executor

        Returns:
            DocTreeRetriever: BM25 retriever
        """
        super().__init__()
        self._top_k = top_k
        self._query_rewrite = query_rewrite
        self._rerank = rerank or DefaultRanker(self._top_k)
        self._keywords_extractor = keywords_extractor
        self._with_content = with_content
        self._show_tree = show_tree
        self._tree_indexes = self._initialize_doc_tree(docs)
        self._executor = executor or ThreadPoolExecutor()

    def get_tree_indexes(self):
        """Get the tree indexes."""
        return self._tree_indexes

    def _retrieve(
        self, query: str, filters: Optional[MetadataFilters] = None
    ) -> List[TreeNode]:
        """Retrieve knowledge chunks.

        Args:
            query (str): query text
            filters: metadata filters.
        Return:
            List[Chunk]: list of chunks
        """
        raise NotImplementedError("DocTreeRetriever does not support retrieval.")

    def _retrieve_with_score(
        self,
        query: str,
        score_threshold: float,
        filters: Optional[MetadataFilters] = None,
    ) -> List[TreeNode]:
        """Retrieve knowledge chunks with score.

        Args:
            query (str): query text
            score_threshold (float): score threshold
            filters: metadata filters.
        Return:
            List[Chunk]: list of chunks with score
        """
        raise NotImplementedError("DocTreeRetriever does not support score retrieval.")

    async def _aretrieve(
        self, query: str, filters: Optional[MetadataFilters] = None
    ) -> List[TreeNode]:
        """Retrieve knowledge chunks.

        Args:
            query (str): query text.
            filters: metadata filters.
        Return:
            List[Chunk]: list of chunks
        """
        keywords = [query]
        if self._keywords_extractor:
            keywords = await self._keywords_extractor.extract(query)
        all_nodes = []
        for keyword in keywords:
            for tree_index in self._tree_indexes:
                retrieve_node = tree_index.search_keywords(tree_index.root, keyword)
                if retrieve_node:
                    # If a match is found, return the corresponding chunks
                    if self._show_tree:
                        logger.info(
                            f"DocTreeIndex Match found in: {retrieve_node.level_text}"
                        )
                        tree_index.display_tree(retrieve_node)
                    all_nodes.append(retrieve_node)
        return all_nodes

    async def _aretrieve_with_score(
        self,
        query: str,
        score_threshold: float,
        filters: Optional[MetadataFilters] = None,
    ) -> List[TreeNode]:
        """Retrieve knowledge chunks with score.

        Args:
            query (str): query text
            score_threshold (float): score threshold
            filters: metadata filters.
        Return:
            List[Chunk]: list of chunks with score
        """
        return await self._aretrieve(query, filters)

    def _initialize_doc_tree(self, docs: List[Document]):
        """Initialize the document tree with docs.

        Args:
            docs (List[Document]): List of docs to initialize the tree with.
        """
        tree_indexes = []
        for doc in docs:
            tree_index = DocTreeIndex()
            for chunk in doc.chunks:
                title = chunk.metadata.get("title") or "title"
                if not self._with_content:
                    tree_index.add_nodes(
                        node_id=chunk.chunk_id,
                        title=title,
                        header1=chunk.metadata.get(HEADER1),
                        header2=chunk.metadata.get(HEADER2),
                        header3=chunk.metadata.get(HEADER3),
                        header4=chunk.metadata.get(HEADER4),
                        header5=chunk.metadata.get(HEADER5),
                        header6=chunk.metadata.get(HEADER6),
                    )
                else:
                    tree_index.add_nodes(
                        node_id=chunk.chunk_id,
                        title=title,
                        header1=chunk.metadata.get(HEADER1),
                        header2=chunk.metadata.get(HEADER2),
                        header3=chunk.metadata.get(HEADER3),
                        header4=chunk.metadata.get(HEADER4),
                        header5=chunk.metadata.get(HEADER5),
                        header6=chunk.metadata.get(HEADER6),
                        content=chunk.content,
                    )
            tree_indexes.append(tree_index)
        return tree_indexes



from dataclasses import dataclass, field
from typing import List, Optional, Type

from dbgpt._private.pydantic import BaseModel, ConfigDict, Field
from dbgpt.core import EmbeddingModelMetadata, Embeddings
from dbgpt.core.interface.parameter import EmbeddingDeployModelParameters
from dbgpt.model.adapter.base import register_embedding_adapter
from dbgpt.util.i18n_utils import _


@dataclass
class SiliconFlowEmbeddingDeployModelParameters(EmbeddingDeployModelParameters):
    """SiliconFlow Embeddings deploy model parameters."""

    provider: str = "proxy/siliconflow"

    api_key: Optional[str] = field(
        default="${env:SILICONFLOW_API_KEY}",
        metadata={
            "help": _("The API key for the rerank API."),
        },
    )
    backend: Optional[str] = field(
        default="BAAI/bge-m3",
        metadata={
            "help": _(
                "The real model name to pass to the provider, default is None. If "
                "backend is None, use name as the real model name."
            ),
        },
    )

    @property
    def real_provider_model_name(self) -> str:
        """Get the real provider model name."""
        return self.backend or self.name


class SiliconFlowEmbeddings(BaseModel, Embeddings):
    """The SiliconFlow embeddings."""

    model_config = ConfigDict(arbitrary_types_allowed=True, protected_namespaces=())
    api_key: Optional[str] = Field(
        default=None, description="The API key for the embeddings API."
    )
    model_name: str = Field(
        default="BAAI/bge-m3", description="The name of the model to use."
    )

    def __init__(self, **kwargs):
        """Initialize the SiliconFlow Embeddings."""
        super().__init__(**kwargs)
        self._api_key = self.api_key

    @classmethod
    def param_class(cls) -> Type[SiliconFlowEmbeddingDeployModelParameters]:
        """Get the parameter class."""
        return SiliconFlowEmbeddingDeployModelParameters

    @classmethod
    def from_parameters(
        cls, parameters: SiliconFlowEmbeddingDeployModelParameters
    ) -> "Embeddings":
        """Create an embedding model from parameters."""
        return cls(
            api_key=parameters.api_key,
            model_name=parameters.real_provider_model_name,
        )

    def embed_documents(
        self, texts: List[str], max_batch_chunks_size=25
    ) -> List[List[float]]:
        """Get the embeddings for a list of texts.

        Args:
            texts (List[str]): A list of texts to get embeddings for.
            max_batch_chunks_size: The max batch size for embedding.

        Returns:
            Embedded texts as List[List[float]], where each inner List[float]
                corresponds to a single input text.
        """
        import requests

        embeddings = []
        headers = {"Authorization": f"Bearer {self._api_key}"}

        for i in range(0, len(texts), max_batch_chunks_size):
            batch_texts = texts[i : i + max_batch_chunks_size]
            response = requests.post(
                url="https://api.siliconflow.cn/v1/embeddings",
                json={"model": self.model_name, "input": batch_texts},
                headers=headers,
            )

            if response.status_code != 200:
                raise RuntimeError(f"Embedding failed: {response.text}")

            # 提取并排序嵌入
            data = response.json()
            batch_embeddings = data["data"]
            sorted_embeddings = sorted(batch_embeddings, key=lambda e: e["index"])
            embeddings.extend([result["embedding"] for result in sorted_embeddings])

        return embeddings

    def embed_query(self, text: str) -> List[float]:
        """Compute query embeddings using a SiliconFlow embedding model.

        Args:
            text: The text to embed.

        Returns:
            Embeddings for the text.
        """
        return self.embed_documents([text])[0]


register_embedding_adapter(
    SiliconFlowEmbeddings,
    supported_models=[
        EmbeddingModelMetadata(
            model="BAAI/bge-m3",
            dimension=1024,
            context_length=8192,
            description=_(
                "The embedding model is provided by SiliconFlow, supporting multiple "
                "languages and high-quality text embeddings."
            ),
        ),
        EmbeddingModelMetadata(
            model="BAAI/bge-large-zh-v1.5",
            dimension=1024,
            context_length=512,
            description=_(
                "The embedding model is provided by SiliconFlow, supporting multiple "
                "languages and high-quality text embeddings."
            ),
        ),
        EmbeddingModelMetadata(
            model="BAAI/bge-large-en-v1.5",
            dimension=1024,
            context_length=512,
            description=_(
                "The embedding model is provided by SiliconFlow, supporting multiple "
                "languages and high-quality text embeddings."
            ),
        ),
    ],
)



"""TuGraph Community Store Adapter."""

import json
import logging
from typing import Any, AsyncGenerator, Dict, Iterator, List, Optional, Tuple, Union

from packaging.version import Version

from dbgpt.storage.graph_store.graph import (
    Direction,
    Edge,
    Graph,
    GraphElemType,
    MemoryGraph,
    Vertex,
)
from dbgpt.storage.knowledge_graph.base import ParagraphChunk
from dbgpt_ext.storage.graph_store.tugraph_store import TuGraphStore
from dbgpt_ext.storage.knowledge_graph.community.base import (
    Community,
    GraphStoreAdapter,
)

logger = logging.getLogger(__name__)


class TuGraphStoreAdapter(GraphStoreAdapter):
    """TuGraph Community Store Adapter."""

    MAX_QUERY_LIMIT = 1000
    MAX_HIERARCHY_LEVEL = 3

    def __init__(self, graph_store: TuGraphStore):
        """Initialize TuGraph Community Store Adapter."""
        super().__init__(graph_store)

        # Create the graph
        self.create_graph(self.graph_store.get_config().name)

    async def discover_communities(self, **kwargs) -> List[str]:
        """Run community discovery with leiden."""
        mg = self.query(
            "CALL db.plugin.callPlugin('CPP',"
            "'leiden','{\"leiden_val\":\"_community_id\"}',60.00,false)"
        )
        result = mg.get_vertex("json_node").get_prop("description")
        community_ids = json.loads(result)["community_id_list"]
        logger.info(f"Discovered {len(community_ids)} communities.")
        return community_ids

    async def get_community(self, community_id: str) -> Community:
        """Get community."""
        query = (
            f"MATCH (n:{self.get_vertex_type()}) WHERE n._community_id = "
            f"'{community_id}' RETURN n"
        )
        edge_query = (
            f"MATCH (n:{self.get_vertex_type()})-"
            f"[r:{self.get_edge_type()}]-"
            f"(m:{self.get_vertex_type()})"
            f"WHERE n._community_id = '{community_id}' RETURN n,r,m"
        )

        all_vertex_graph = self.query(query)
        all_edge_graph = self.query(edge_query)
        all_graph = MemoryGraph()
        for vertex in all_vertex_graph.vertices():
            all_graph.upsert_vertex(vertex)
        for edge in all_edge_graph.edges():
            all_graph.append_edge(edge)

        return Community(id=community_id, data=all_graph)

    @property
    def graph_store(self) -> TuGraphStore:
        """Get the graph store."""
        return self._graph_store  # type: ignore[return-value]

    def get_graph_config(self):
        """Get the graph store config."""
        return self.graph_store.get_config()

    def get_vertex_type(self) -> str:
        """Get the vertex type."""
        return GraphElemType.ENTITY.value

    def get_edge_type(self) -> str:
        """Get the edge type."""
        return GraphElemType.RELATION.value

    def get_triplets(self, subj: str) -> List[Tuple[str, str]]:
        """Get triplets."""
        triplet_query = (
            f"MATCH (n1:{GraphElemType.ENTITY.value})-[r]->(n2:"
            f"{GraphElemType.ENTITY.value}) "
            f'WHERE n1.id = "{subj}" RETURN r.id as rel, n2.id as obj;'
        )
        data = self.graph_store.conn.run(triplet_query)
        return [(record["rel"], record["obj"]) for record in data]

    def get_document_vertex(self, doc_name: str) -> Vertex:
        """Get the document vertex in the graph."""
        gql = f"""MATCH (n) WHERE n.id = {doc_name} RETURN n"""
        graph = self.query(gql)
        vertex = graph.get_vertex(doc_name)
        return vertex

    def get_schema(self, refresh: bool = False) -> str:
        """Get the schema of the graph store."""
        query = "CALL dbms.graph.getGraphSchema()"
        data = self.graph_store.conn.run(query=query)
        schema = data[0]["schema"]
        return schema

    def get_full_graph(self, limit: Optional[int] = None) -> Graph:
        """Get full graph."""
        if not limit:
            limit = self.MAX_QUERY_LIMIT
        if limit <= 0:
            raise ValueError("Limit must be greater than 0.")
        graph_result = self.query(
            f"MATCH (n)-[r]-(m) RETURN n,r,m LIMIT {limit}",
            white_list=["_community_id"],
        )
        full_graph = MemoryGraph()
        for vertex in graph_result.vertices():
            full_graph.upsert_vertex(vertex)
        for edge in graph_result.edges():
            full_graph.append_edge(edge)
        return full_graph

    def upsert_entities(self, entities: Iterator[Vertex]) -> None:
        """Upsert entities."""
        enable_similarity_search = self.graph_store.enable_similarity_search
        entity_list = [
            {
                "id": self._escape_quotes(entity.vid),
                "name": self._escape_quotes(entity.name),
                "description": self._escape_quotes(entity.get_prop("description"))
                or "",
                "_document_id": "0",
                "_chunk_id": "0",
                "_community_id": "0",
                **(
                    {"_embedding": entity.get_prop("_embedding")}
                    if enable_similarity_search
                    else {}
                ),
            }
            for entity in entities
        ]
        entity_query = (
            f"CALL db.upsertVertex("
            f'"{GraphElemType.ENTITY.value}", '
            f"[{self._convert_dict_to_str(entity_list)}])"
        )

        # If similarity search enabled, then ready to create vector index
        if enable_similarity_search:
            # Check wheather the vector index exist
            check_entity_vector_query = (
                "CALL db.showVertexVectorIndex() "
                "YIELD label_name, field_name "
                f"WHERE label_name = '{GraphElemType.ENTITY.value}' "
                "AND field_name = '_embedding' "
                "RETURN label_name"
            )
            # If not exist, then create vector index
            if self.query(check_entity_vector_query).vertex_count == 0:
                # Get the dimension
                dimension = len(entity_list[0].get("_embedding", []))
                # Then create index
                create_vector_index_query = (
                    "CALL db.addVertexVectorIndex("
                    f'"{GraphElemType.ENTITY.value}", "_embedding", '
                    f"{{dimension: {dimension}}})"
                )
                self.graph_store.conn.run(query=create_vector_index_query)

        self.graph_store.conn.run(query=entity_query)

    def upsert_edge(
        self, edges: Iterator[Edge], edge_type: str, src_type: str, dst_type: str
    ) -> None:
        """Upsert edges."""
        edge_list = [
            {
                "sid": self._escape_quotes(edge.sid),
                "tid": self._escape_quotes(edge.tid),
                "id": self._escape_quotes(edge.name),
                "name": self._escape_quotes(edge.name),
                "description": self._escape_quotes(edge.get_prop("description")) or "",
                "_chunk_id": self._escape_quotes(edge.get_prop("_chunk_id")) or "",
            }
            for edge in edges
        ]
        relation_query = f"""CALL db.upsertEdge("{edge_type}",
            {{type:"{src_type}", key:"sid"}},
            {{type:"{dst_type}", key:"tid"}},
            [{self._convert_dict_to_str(edge_list)}])"""
        self.graph_store.conn.run(query=relation_query)

    def upsert_chunks(self, chunks: Iterator[Union[Vertex, ParagraphChunk]]) -> None:
        """Upsert chunks."""
        enable_similarity_search = self.graph_store.enable_similarity_search
        chunk_list: List[Dict[str, Union[str, List[float]]]] = [
            {
                "id": self._escape_quotes(chunk.chunk_id),
                "name": self._escape_quotes(chunk.chunk_name),
                "content": self._escape_quotes(chunk.content),
                **(
                    {"_embedding": chunk.embedding}
                    if enable_similarity_search and chunk.embedding
                    else {}
                ),
            }
            if isinstance(chunk, ParagraphChunk)
            else {
                "id": self._escape_quotes(chunk.vid),
                "name": self._escape_quotes(chunk.name),
                "content": self._escape_quotes(chunk.get_prop("content")),
                **(
                    {"_embedding": chunk.get_prop("_embedding")}
                    if enable_similarity_search
                    else {}
                ),
            }
            for chunk in chunks
        ]
        if len(chunk_list) == 0:
            return

        chunk_query = (
            f"CALL db.upsertVertex("
            f'"{GraphElemType.CHUNK.value}", '
            f"[{self._convert_dict_to_str(chunk_list)}])"
        )

        # If similarity search enabled, then ready to create vector index
        if enable_similarity_search:
            # Check wheather the vector index exist
            check_chunk_vector_query = (
                "CALL db.showVertexVectorIndex() "
                "YIELD label_name, field_name "
                f"WHERE label_name = '{GraphElemType.CHUNK.value}' "
                "AND field_name = '_embedding' "
                "RETURN label_name"
            )
            # If not exist, then create vector index
            if self.query(check_chunk_vector_query).vertex_count == 0:
                # Get the dimension
                dimension = len(chunk_list[0].get("_embedding", []))
                # Then create index
                create_vector_index_query = (
                    "CALL db.addVertexVectorIndex("
                    f'"{GraphElemType.CHUNK.value}", "_embedding", '
                    f"{{dimension: {dimension}}})"
                )
                self.graph_store.conn.run(query=create_vector_index_query)

        self.graph_store.conn.run(query=chunk_query)

    def upsert_documents(
        self, documents: Iterator[Union[Vertex, ParagraphChunk]]
    ) -> None:
        """Upsert documents."""
        document_list = [
            (
                {
                    "id": self._escape_quotes(document.chunk_id),
                    "name": self._escape_quotes(document.chunk_name),
                    "content": "",
                }
                if isinstance(document, ParagraphChunk)
                else {
                    "id": self._escape_quotes(document.vid),
                    "name": self._escape_quotes(document.name),
                    "content": "",
                }
            )
            for document in documents
        ]

        document_query = (
            "CALL db.upsertVertex("
            f'"{GraphElemType.DOCUMENT.value}", '
            f"[{self._convert_dict_to_str(document_list)}])"
        )
        self.graph_store.conn.run(query=document_query)

    def insert_triplet(self, subj: str, rel: str, obj: str) -> None:
        """Add triplet."""
        subj_escaped = subj.replace("'", "\\'").replace('"', '\\"')
        rel_escaped = rel.replace("'", "\\'").replace('"', '\\"')
        obj_escaped = obj.replace("'", "\\'").replace('"', '\\"')

        vertex_query = f"""CALL db.upsertVertex(
            '{GraphElemType.ENTITY.value}',
            [{{id:'{subj_escaped}',name:'{subj_escaped}'}},
            {{id:'{obj_escaped}',name:'{obj_escaped}'}}])"""
        edge_query = f"""CALL db.upsertEdge(
            '{GraphElemType.RELATION.value}',
            {{type:"{GraphElemType.ENTITY.value}",key:"sid"}},
            {{type:"{GraphElemType.ENTITY.value}", key:"tid"}},
            [{{sid:"{subj_escaped}",
            tid: "{obj_escaped}",
            id:"{rel_escaped}",
            name: "{rel_escaped}"}}])"""

        self.graph_store.conn.run(query=vertex_query)
        self.graph_store.conn.run(query=edge_query)

    def upsert_graph(self, graph: Graph) -> None:
        """Add graph to the graph store.

        Args:
            graph (Graph): The graph to be added.
        """
        # Get the iterators of all the vertices and the edges from the graph
        documents: Iterator[Vertex] = graph.vertices(
            filter_fn=lambda x: (
                x.get_prop("vertex_type") == GraphElemType.DOCUMENT.value
            )
        )
        chunks: Iterator[Vertex] = graph.vertices(
            filter_fn=lambda x: x.get_prop("vertex_type") == GraphElemType.CHUNK.value
        )
        entities: Iterator[Vertex] = graph.vertices(
            filter_fn=lambda x: x.get_prop("vertex_type") == GraphElemType.ENTITY.value
        )
        doc_include_chunk: Iterator[Edge] = graph.edges(
            filter_fn=lambda x: (
                x.get_prop("edge_type") == GraphElemType.DOCUMENT_INCLUDE_CHUNK.value
            )
        )
        chunk_include_chunk: Iterator[Edge] = graph.edges(
            filter_fn=lambda x: (
                x.get_prop("edge_type") == GraphElemType.CHUNK_INCLUDE_CHUNK.value
            )
        )
        chunk_include_entity: Iterator[Edge] = graph.edges(
            filter_fn=lambda x: (
                x.get_prop("edge_type") == GraphElemType.CHUNK_INCLUDE_ENTITY.value
            )
        )
        chunk_next_chunk: Iterator[Edge] = graph.edges(
            filter_fn=lambda x: (
                x.get_prop("edge_type") == GraphElemType.CHUNK_NEXT_CHUNK.value
            )
        )
        relation: Iterator[Edge] = graph.edges(
            filter_fn=lambda x: x.get_prop("edge_type") == GraphElemType.RELATION.value
        )

        # Upsert the vertices and the edges to the graph store
        self.upsert_entities(entities)
        self.upsert_chunks(chunks)
        self.upsert_documents(documents)
        self.upsert_edge(
            doc_include_chunk,
            GraphElemType.INCLUDE.value,
            GraphElemType.DOCUMENT.value,
            GraphElemType.CHUNK.value,
        )
        self.upsert_edge(
            chunk_include_chunk,
            GraphElemType.INCLUDE.value,
            GraphElemType.CHUNK.value,
            GraphElemType.CHUNK.value,
        )
        self.upsert_edge(
            chunk_include_entity,
            GraphElemType.INCLUDE.value,
            GraphElemType.CHUNK.value,
            GraphElemType.ENTITY.value,
        )
        self.upsert_edge(
            chunk_next_chunk,
            GraphElemType.NEXT.value,
            GraphElemType.CHUNK.value,
            GraphElemType.CHUNK.value,
        )
        self.upsert_edge(
            relation,
            GraphElemType.RELATION.value,
            GraphElemType.ENTITY.value,
            GraphElemType.ENTITY.value,
        )

    def delete_document(self, chunk_id: str) -> None:
        """Delete document in the graph."""
        chunkids_list = [uuid.strip() for uuid in chunk_id.split(",")]
        del_chunk_gql = (
            f"MATCH(m:{GraphElemType.DOCUMENT.value})-[r]->"
            f"(n:{GraphElemType.CHUNK.value}) WHERE n.id IN {chunkids_list} DELETE n"
        )
        del_relation_gql = (
            f"MATCH(m:{GraphElemType.ENTITY.value})-[r:"
            f"{GraphElemType.RELATION.value}]-(n:{GraphElemType.ENTITY.value}) "
            f"WHERE r._chunk_id IN {chunkids_list} DELETE r"
        )
        delete_only_vertex = "MATCH (n) WHERE NOT EXISTS((n)-[]-()) DELETE n"
        self.graph_store.conn.run(del_chunk_gql)
        self.graph_store.conn.run(del_relation_gql)
        self.graph_store.conn.run(delete_only_vertex)

    def delete_triplet(self, sub: str, rel: str, obj: str) -> None:
        """Delete triplet."""
        del_query = (
            f"MATCH (n1:{GraphElemType.ENTITY.value} {{id:'{sub}'}})"
            f"-[r:{GraphElemType.RELATION.value} {{id:'{rel}'}}]->"
            f"(n2:{GraphElemType.ENTITY.value} {{id:'{obj}'}}) DELETE n1,n2,r"
        )
        self.graph_store.conn.run(query=del_query)

    def drop(self):
        """Delete Graph."""
        self.graph_store.conn.delete_graph(self.get_graph_config().name)

    def create_graph(self, graph_name: str):
        """Create a graph."""
        if not self.graph_store.conn.create_graph(graph_name=graph_name):
            return

        # Compatibility check
        enable_similarity_search = self.graph_store.enable_similarity_search

        dbms_system_info = self.graph_store.conn.get_system_info()
        lgraph_version = dbms_system_info["lgraph_version"]
        similarity_search_compatible = Version(lgraph_version) >= Version("4.5.1")

        if enable_similarity_search and not similarity_search_compatible:
            raise Exception(
                "TuGraph 4.5.0 and below does not support similarity search."
            )

        # Create the graph schema
        def _format_graph_property_schema(
            name: str,
            type: str = "STRING",
            optional: bool = False,
            index: Optional[bool] = None,
            **kwargs,
        ) -> Dict[str, str | bool]:
            """Format the property for TuGraph.

            Args:
                name: The name of the property.
                type: The type of the property.
                optional: The optional of the property.
                index: The index of the property.
                kwargs: Additional keyword arguments.

            Returns:
                The formatted property.
            """
            property: Dict[str, str | bool] = {
                "name": name,
                "type": type,
                "optional": optional,
            }

            if index is not None:
                property["index"] = index

            # Add any additional keyword arguments to the property dictionary
            property.update(kwargs)
            return property

        # Create the graph label for document vertex
        document_proerties: List[Dict[str, Union[str, bool]]] = [
            _format_graph_property_schema("id", "STRING", False),
            _format_graph_property_schema("name", "STRING", False),
            _format_graph_property_schema("_community_id", "STRING", True, True),
        ]
        self.create_graph_label(
            graph_elem_type=GraphElemType.DOCUMENT, graph_properties=document_proerties
        )

        # Create the graph label for chunk vertex
        chunk_proerties: List[Dict[str, Union[str, bool]]] = [
            _format_graph_property_schema("id", "STRING", False),
            _format_graph_property_schema("name", "STRING", False),
            _format_graph_property_schema("_community_id", "STRING", True, True),
            _format_graph_property_schema("content", "STRING", True, True),
        ]
        if enable_similarity_search:
            chunk_proerties.append(
                _format_graph_property_schema("_embedding", "FLOAT_VECTOR", True, False)
            )
        self.create_graph_label(
            graph_elem_type=GraphElemType.CHUNK, graph_properties=chunk_proerties
        )

        # Create the graph label for entity vertex
        vertex_proerties: List[Dict[str, Union[str, bool]]] = [
            _format_graph_property_schema("id", "STRING", False),
            _format_graph_property_schema("name", "STRING", False),
            _format_graph_property_schema("_community_id", "STRING", True, True),
            _format_graph_property_schema("description", "STRING", True, True),
        ]
        if enable_similarity_search:
            vertex_proerties.append(
                _format_graph_property_schema(
                    "_embedding", "FLOAT_VECTOR", True, False
                ),
            )
        self.create_graph_label(
            graph_elem_type=GraphElemType.ENTITY, graph_properties=vertex_proerties
        )

        # Create the graph label for relation edge
        edge_proerties: List[Dict[str, Union[str, bool]]] = [
            _format_graph_property_schema("id", "STRING", False),
            _format_graph_property_schema("name", "STRING", False),
            _format_graph_property_schema("_chunk_id", "STRING", True, True),
            _format_graph_property_schema("description", "STRING", True, True),
        ]
        self.create_graph_label(
            graph_elem_type=GraphElemType.RELATION, graph_properties=edge_proerties
        )

        # Create the graph label for include edge
        include_proerties: List[Dict[str, Union[str, bool]]] = [
            _format_graph_property_schema("id", "STRING", False),
            _format_graph_property_schema("name", "STRING", False),
            _format_graph_property_schema("description", "STRING", True),
        ]
        self.create_graph_label(
            graph_elem_type=GraphElemType.INCLUDE, graph_properties=include_proerties
        )

        # Create the graph label for next edge
        next_proerties: List[Dict[str, Union[str, bool]]] = [
            _format_graph_property_schema("id", "STRING", False),
            _format_graph_property_schema("name", "STRING", False),
            _format_graph_property_schema("description", "STRING", True),
        ]
        self.create_graph_label(
            graph_elem_type=GraphElemType.NEXT, graph_properties=next_proerties
        )

        if self.graph_store.enable_summary:
            self.graph_store._upload_plugin()

    def create_graph_label(
        self,
        graph_elem_type: GraphElemType,
        graph_properties: List[Dict[str, Union[str, bool]]],
    ) -> None:
        """Create a graph label.

        The graph label is used to identify and distinguish different types of nodes
        (vertices) and edges in the graph.
        """
        if graph_elem_type.is_vertex():  # vertex
            vertex_meta = json.dumps(
                {
                    "label": graph_elem_type.value,
                    "type": "VERTEX",
                    "primary": "id",
                    "properties": graph_properties,
                }
            )
            gql = f"""CALL db.createVertexLabelByJson('{vertex_meta}')"""
        else:  # edge

            def edge_direction(graph_elem_type: GraphElemType) -> List[List[str]]:
                """Define the edge direction.

                `include` edge: document -> chunk, chunk -> entity
                `next` edge: chunk -> chunk
                `relation` edge: entity -> entity
                """
                if graph_elem_type.is_vertex():
                    raise ValueError("The graph element type must be an edge.")
                if graph_elem_type == GraphElemType.INCLUDE:
                    return [
                        [GraphElemType.DOCUMENT.value, GraphElemType.CHUNK.value],
                        [GraphElemType.CHUNK.value, GraphElemType.ENTITY.value],
                        [GraphElemType.CHUNK.value, GraphElemType.CHUNK.value],
                    ]
                elif graph_elem_type == GraphElemType.NEXT:
                    return [[GraphElemType.CHUNK.value, GraphElemType.CHUNK.value]]
                elif graph_elem_type == GraphElemType.RELATION:
                    return [[GraphElemType.ENTITY.value, GraphElemType.ENTITY.value]]
                else:
                    raise ValueError("Invalid graph element type.")

            edge_meta = json.dumps(
                {
                    "label": graph_elem_type.value,
                    "type": "EDGE",
                    "constraints": edge_direction(graph_elem_type),
                    "properties": graph_properties,
                }
            )
            gql = f"""CALL db.createEdgeLabelByJson('{edge_meta}')"""

        self.graph_store.conn.run(gql)

    def truncate(self):
        """Truncate Graph."""
        gql = "MATCH (n) DELETE n"
        self.graph_store.conn.run(gql)

    def check_label(self, graph_elem_type: GraphElemType) -> bool:
        """Check if the label exists in the graph.

        Args:
            graph_elem_type (GraphElemType): The type of the graph element.

        Returns:
            True if the label exists in the specified graph element type, otherwise
            False.
        """
        tables = self.graph_store.conn.get_table_names()

        return graph_elem_type.value in tables

    def explore_trigraph(
        self,
        subs: Union[List[str], List[List[float]]],
        topk: Optional[int] = None,
        score_threshold: Optional[float] = None,
        direct: Direction = Direction.BOTH,
        depth: int = 3,
        fan: Optional[int] = None,
        limit: Optional[int] = None,
    ) -> MemoryGraph:
        """Explore the graph from given subjects up to a depth.

        Args:
            subs (Union[List[str], List[List[float]]): The list of the subjects
                (keywords or embedding vectors).
            topk (Optional[int]): The number of the top similar entities.
            score_threshold (Optional[float]): The threshold of the similarity score.
            direct (Direction): The direction of the graph that will be explored.
            depth (int): The depth of the graph that will be explored.
            fan (Optional[int]): Not used.
            limit (Optional[int]): The limit number of the queried entities.

        Returns:
            MemoryGraph: The triplet graph that includes the entities and the relations.
        """
        if not subs:
            return MemoryGraph()

        if depth <= 0:
            depth = 3
        depth_string = f"1..{depth}"

        if limit is None:
            limit_string = ""
        else:
            limit_string = f"LIMIT {limit}"

        if direct.name == "OUT":
            rel = f"-[r:{GraphElemType.RELATION.value}*{depth_string}]->"
        elif direct.name == "IN":
            rel = f"<-[r:{GraphElemType.RELATION.value}*{depth_string}]-"
        else:
            rel = f"-[r:{GraphElemType.RELATION.value}*{depth_string}]-"

        enable_similarity_search = self.graph_store.enable_similarity_search

        if enable_similarity_search:
            # If enable similarity search, using knn-search to get the id
            similar_entities = []
            # Get the vector from vectors
            # Then do knn-search for each vectors by using TuGraph
            for vector in subs:
                similarity_retrieval_query = (
                    "CALL db.vertexVectorKnnSearch("
                    f"'{GraphElemType.ENTITY.value}','_embedding', {vector}, "
                    f"{{top_k:{topk}}}) YIELD node "
                    f"WHERE node.distance < {score_threshold} "
                    "RETURN node.id AS id;"
                )
                # Merge the result for each knn-search result
                similar_entities.extend(
                    self.graph_store.conn.run(query=similarity_retrieval_query)
                )
            # Get the id from the retrieved entities
            ids = [(record["id"]) for record in similar_entities]
            conditional_statement = f"WHERE n.id IN {ids} "
        else:
            conditional_statement = (
                f"WHERE n.id IN {[self._escape_quotes(str(sub)) for sub in subs]} "
            )

        # Multi-hop search
        query = (
            f"MATCH p=(n:{GraphElemType.ENTITY.value})"
            f"{rel}(m:{GraphElemType.ENTITY.value}) "
            f"{conditional_statement}"
            f"RETURN p {limit_string}"
        )

        return self.query(query=query, white_list=["description"])

    def explore_docgraph_with_entities(
        self,
        subs: List[str],
        topk: Optional[int] = None,
        score_threshold: Optional[float] = None,
        direct: Direction = Direction.BOTH,
        depth: int = 3,
        fan: Optional[int] = None,
        limit: Optional[int] = None,
    ) -> MemoryGraph:
        """Explore the graph from given subjects up to a depth.

        Args:
            subs (List[str]): The list of the entities.
            topk (Optional[int]): The number of the top similar chunks.
            score_threshold (Optional[float]): The threshold of the similarity score.
            direct (Direction): The direction of the graph that will be explored.
            depth (int): The depth of the graph that will be explored.
            fan (Optional[int]): Not used.
            limit (Optional[int]): The limit number of the queried chunks.

        Returns:
            MemoryGraph: The document graph that includes the leaf chunks that
                connect to the entities, the chains from documents to the leaf chunks,
                and the chain from documents to chunks.
        """
        if len(subs) == 0:
            return MemoryGraph()

        if depth <= 0:
            depth = 3
        depth_string = f"1..{depth}"

        if limit is None:
            limit_string = ""
        else:
            limit_string = f"LIMIT {limit}"

        # If there exists the entities in the graph, return the graph that
        # includes the leaf chunks that connect to the entities, the chains from
        # documents to the leaf chunks, and the chain from documents to chunks;
        # document -> chunk -> chunk -> ... -> leaf chunk -> (entity)
        #
        # If not, return the graph that includes the chains from documents to chunks
        # that contain the subs (keywords).
        # document -> chunk -> chunk -> ... -> leaf chunk (that contains the subs)
        #
        # And only the leaf chunks contain the content, and the other chunks do not
        # contain any properties except the id, name.

        graph = MemoryGraph()

        # Query the leaf chunks in the chain from documents to chunks
        conditional_statement = (
            f"WHERE m.name IN {[self._escape_quotes(sub) for sub in subs]} "
        )
        leaf_chunk_query = (
            f"MATCH p=(n:{GraphElemType.CHUNK.value})-"
            f"[r:{GraphElemType.INCLUDE.value}]->"
            f"(m:{GraphElemType.ENTITY.value})"
            f"{conditional_statement} "
            "RETURN n"
        )
        graph_of_leaf_chunks = self.query(
            query=leaf_chunk_query, white_list=["content"]
        )

        # Query the chain from documents to chunks,
        # document -> chunk -> ... ->  leaf_chunks
        chunk_names = [
            self._escape_quotes(vertex.name)
            for vertex in graph_of_leaf_chunks.vertices()
        ]
        chain_query = (
            f"MATCH p=(n:{GraphElemType.DOCUMENT.value})-"
            f"[:{GraphElemType.INCLUDE.value}*{depth_string}]->"
            f"(m:{GraphElemType.CHUNK.value})"
            f"WHERE m.name IN {chunk_names} "
            "RETURN p"
        )
        # Filter all the properties by with_list
        graph.upsert_graph(self.query(query=chain_query, white_list=[""]))

        # The number of leaf chunks caompared to the `limit`
        if not limit or len(chunk_names) <= limit:
            graph.upsert_graph(graph_of_leaf_chunks)
        else:
            limited_leaf_chunk_query = leaf_chunk_query + f" {limit_string}"
            graph.upsert_graph(
                self.query(query=limited_leaf_chunk_query, white_list=["content"])
            )

        return graph

    def explore_docgraph_without_entities(
        self,
        subs: Union[List[str], List[List[float]]],
        topk: Optional[int] = None,
        score_threshold: Optional[float] = None,
        direct: Direction = Direction.BOTH,
        depth: int = 3,
        fan: Optional[int] = None,
        limit: Optional[int] = None,
    ) -> MemoryGraph:
        """Explore the graph from given subjects up to a depth.

        Args:
            subs (Union[List[str], List[List[float]]): The list of the subjects
                (keywords or embedding vectors).
            topk (Optional[int]): The number of the top similar chunks.
            score_threshold (Optional[float]): The threshold of the similarity score.
            direct (Direction): The direction of the graph that will be explored.
            depth (int): The depth of the graph that will be explored.
            fan (Optional[int]): Not used.
            limit (Optional[int]): The limit number of the queried chunks.

        Returns:
            MemoryGraph: The document graph that includes the chains from documents
                to chunks that contain the subs (keywords) or similar chunks
                (embedding vectors).
        """
        if len(subs) == 0:
            return MemoryGraph()

        if depth <= 0:
            depth = 3
        depth_string = f"1..{depth}"

        if limit is None:
            limit_string = ""
        else:
            limit_string = f"LIMIT {limit}"

        enable_similarity_search = self.graph_store.enable_similarity_search

        # If there exists the entities in the graph, return the graph that
        # includes the leaf chunks that connect to the entities, the chains from
        # documents to the leaf chunks, and the chain from documents to chunks;
        # document -> chunk -> chunk -> ... -> leaf chunk -> (entity)
        #
        # If not, return the graph that includes the chains from documents to chunks
        # that contain the subs (keywords).
        # document -> chunk -> chunk -> ... -> leaf chunk (that contains the subs)
        #
        # And only the leaf chunks contain the content, and the other chunks do not
        # contain any properties except the id, name.

        graph = MemoryGraph()

        if enable_similarity_search:
            similar_chunks = []
            for vector in subs:
                similarity_retrieval_query = (
                    "CALL db.vertexVectorKnnSearch("
                    f"'{GraphElemType.CHUNK.value}','_embedding', {vector}, "
                    f"{{top_k:{topk}}}) YIELD node "
                    f"WHERE node.distance < {score_threshold} "
                    "RETURN node.name AS name"
                )
                similar_chunks.extend(
                    self.graph_store.conn.run(query=similarity_retrieval_query)
                )
            names = [(record["name"]) for record in similar_chunks]
            _subs_condition = " OR ".join(
                [f"m.content CONTAINS '{name}'" for name in names]
            )
        else:
            _subs_condition = " OR ".join(
                [
                    f"m.content CONTAINS '{self._escape_quotes(str(sub))}'"
                    for sub in subs
                ]
            )

        # Query the chain from documents to chunks,
        # document -> chunk -> chunk -> chunk -> ... -> chunk
        chain_query = (
            f"MATCH p=(n:{GraphElemType.DOCUMENT.value})-"
            f"[r:{GraphElemType.INCLUDE.value}*{depth_string}]->"
            f"(m:{GraphElemType.CHUNK.value})"
            f"WHERE {_subs_condition}"
            "RETURN p"
        )
        # Filter all the properties by with_list
        graph.upsert_graph(self.query(query=chain_query, white_list=[""]))

        # Query the leaf chunks in the chain from documents to chunks
        leaf_chunk_query = (
            f"MATCH p=(n:{GraphElemType.DOCUMENT.value})-"
            f"[r:{GraphElemType.INCLUDE.value}*{depth_string}]->"
            f"(m:{GraphElemType.CHUNK.value})"
            f"WHERE {_subs_condition}"
            f"RETURN m {limit_string}"
        )
        graph.upsert_graph(self.query(query=leaf_chunk_query, white_list=["content"]))

        return graph

    def query(self, query: str, **kwargs) -> MemoryGraph:
        """Execute a query on graph.

        white_list: List[str] = kwargs.get("white_list", []),
        which contains the white list of properties that start with '_',
        filters all the properties that start with '_' and are not in the white list.

        white_list: List of properties to keep
                    - If []: Keep default properties (those not starting with '_'
                        and not in ['id', 'name'])
                    - If list of strings: Keep default properties
                        (those not starting with '_' and not in ['id', 'name'])
                        and properties in white_list
        """
        query_result = self.graph_store.conn.run(query=query)
        white_list: List[str] = kwargs.get("white_list", [])
        vertices, edges = self._get_nodes_edges_from_queried_data(
            query_result, white_list
        )

        mg = MemoryGraph()
        for vertex in vertices:
            mg.upsert_vertex(vertex)
        for edge in edges:
            mg.append_edge(edge)
        return mg

    # type: ignore[override]
    # mypy: ignore-errors
    async def stream_query(  # type: ignore[override]
        self,
        query: str,
        **kwargs,
    ) -> AsyncGenerator[Graph, None]:
        """Execute a stream query."""
        from neo4j import graph

        async for record in self.graph_store.conn.run_stream(query):  # type: ignore
            mg = MemoryGraph()
            for key in record.keys():
                value = record[key]
                if isinstance(value, graph.Node):
                    node_id = value._properties["id"]
                    description = value._properties["description"]
                    vertex = Vertex(vid=node_id, name=node_id, description=description)
                    mg.upsert_vertex(vertex)
                elif isinstance(value, graph.Relationship):
                    edge_nodes = value.nodes
                    prop_id = value._properties["id"]
                    assert edge_nodes and edge_nodes[0] and edge_nodes[1]
                    src_id = edge_nodes[0]._properties["id"]
                    dst_id = edge_nodes[1]._properties["id"]
                    description = value._properties["description"]
                    edge = Edge(
                        sid=src_id, tid=dst_id, name=prop_id, description=description
                    )
                    mg.append_edge(edge)
                elif isinstance(value, graph.Path):
                    nodes = list(record["p"].nodes)
                    rels = list(record["p"].relationships)
                    formatted_path = []
                    for i in range(len(nodes)):
                        formatted_path.append(
                            {
                                "id": nodes[i]._properties["id"],
                                "description": nodes[i]._properties["description"],
                            }
                        )
                        if i < len(rels):
                            formatted_path.append(
                                {
                                    "id": rels[i]._properties["id"],
                                    "description": rels[i]._properties["description"],
                                }
                            )
                    for i in range(0, len(formatted_path), 2):
                        mg.upsert_vertex(
                            Vertex(
                                vid=formatted_path[i]["id"],
                                name=formatted_path[i]["id"],
                                description=formatted_path[i]["description"],
                            )
                        )
                        if i + 2 < len(formatted_path):
                            mg.append_edge(
                                Edge(
                                    sid=formatted_path[i]["id"],
                                    tid=formatted_path[i + 2]["id"],
                                    name=formatted_path[i + 1]["id"],
                                    description=formatted_path[i + 1]["description"],
                                )
                            )
                else:
                    vertex = Vertex(
                        vid="json_node", name="json_node", description=value
                    )
                    mg.upsert_vertex(vertex)
            yield mg

    def _get_nodes_edges_from_queried_data(
        self,
        data: List[Dict[str, Any]],
        white_prop_list: List[str],
    ) -> Tuple[List[Vertex], List[Edge]]:
        """Format the query data.

        Args:
            data: The data to be formatted.
            white_prop_list: The white list of properties.

        Returns:
            Tuple[List[Vertex], List[Edge]]: The formatted vertices and edges.
        """
        vertex_list: List[Vertex] = []
        edge_list: List[Edge] = []

        # Remove id, src_id, dst_id and name from the white list
        # to avoid duplication in the initialisation of the vertex and edge
        _white_list = [
            prop
            for prop in white_prop_list
            if prop not in ["id", "src_id", "dst_id", "name"]
        ]

        from neo4j import graph

        def filter_properties(
            properties: dict[str, Any], white_list: List[str]
        ) -> Dict[str, Any]:
            """Filter the properties.

            It will remove the properties that are not in the white list.
            The expected propertities are:
                entity_properties = ["id", "name", "description", "_document_id",
                                        "_chunk_id", "_community_id"]
                edge_properties = ["id", "name", "description", "_chunk_id"]
            Args:
                properties: Dictionary of properties to filter
                white_list: List of properties to keep
                    - If [""]: Keep default properties (those not starting with '_'
                        and not in ['id', 'name'])
                    - If list of strings: Keep only properties in white_list
            """
            return {
                key: value
                for key, value in properties.items()
                if (
                    (not key.startswith("_") and key not in ["id", "name"])
                    or (key in white_list)
                )
            }

        # Parse the data to nodes and relationships
        for record in data:
            for value in record.values():
                if isinstance(value, graph.Node):
                    assert value._properties.get("id")
                    vertex = Vertex(
                        vid=value._properties.get("id", ""),
                        name=value._properties.get("name"),
                        **filter_properties(value._properties, _white_list),
                    )
                    if vertex not in vertex_list:
                        # TODO: Do we really need to check it every time?
                        vertex_list.append(vertex)
                elif isinstance(value, graph.Relationship):
                    for node in value.nodes:  # num of nodes is 2
                        assert node and node._properties
                        vertex = Vertex(
                            vid=node._properties.get("id", ""),
                            name=node._properties.get("name"),
                            **filter_properties(node._properties, _white_list),
                        )
                        if vertex not in vertex_list:
                            vertex_list.append(vertex)

                        assert value.nodes and value.nodes[0] and value.nodes[1]
                        edge = Edge(
                            sid=value.nodes[0]._properties.get("id", ""),
                            tid=value.nodes[1]._properties.get("id", ""),
                            name=value._properties.get("name", ""),
                            **filter_properties(value._properties, _white_list),
                        )
                        if edge not in edge_list:
                            edge_list.append(edge)
                elif isinstance(value, graph.Path):
                    for rel in value.relationships:
                        for node in rel.nodes:  # num of nodes is 2
                            assert node and node._properties
                            vertex = Vertex(
                                vid=node._properties.get("id", ""),
                                name=node._properties.get("name"),
                                **filter_properties(node._properties, _white_list),
                            )
                            if vertex not in vertex_list:
                                vertex_list.append(vertex)

                            assert rel.nodes and rel.nodes[0] and rel.nodes[1]
                            edge = Edge(
                                sid=rel.nodes[0]._properties.get("id", ""),
                                tid=rel.nodes[1]._properties.get("id", ""),
                                name=rel._properties.get("name", ""),
                                **filter_properties(rel._properties, _white_list),
                            )
                            if edge not in edge_list:
                                edge_list.append(edge)

                else:  # json_node
                    vertex = Vertex(
                        vid="json_node",
                        name="json_node",
                        **filter_properties({"description": value}, _white_list),
                    )
                    if vertex not in vertex_list:
                        vertex_list.append(vertex)
        return vertex_list, edge_list

    def _convert_dict_to_str(self, entity_list: List[Dict[str, Any]]) -> str:
        """Convert a list of entities to a formatted string representation."""
        formatted_nodes = [
            "{"
            + ", ".join(
                f'{k}: "{v}"' if isinstance(v, str) else f"{k}: {v}"
                for k, v in node.items()
            )
            + "}"
            for node in entity_list
        ]
        return f"""{", ".join(formatted_nodes)}"""

    def _escape_quotes(self, value: str) -> str:
        """Escape single and double quotes in a string for queries."""
        if value is not None:
            return value.replace("'", "").replace('"', "")

    def upsert_doc_include_chunk(
        self,
        chunk: ParagraphChunk,
    ) -> None:
        """Convert chunk to document include chunk."""
        assert chunk.chunk_parent_id and chunk.chunk_parent_name, (
            "Chunk parent ID and name are required (document_include_chunk)"
        )

        edge = Edge(
            sid=chunk.chunk_parent_id,
            tid=chunk.chunk_id,
            name=GraphElemType.INCLUDE.value,
            edge_type=GraphElemType.DOCUMENT_INCLUDE_CHUNK.value,
        )

        self.upsert_edge(
            edges=iter([edge]),
            edge_type=GraphElemType.INCLUDE.value,
            src_type=GraphElemType.DOCUMENT.value,
            dst_type=GraphElemType.CHUNK.value,
        )

    def upsert_chunk_include_chunk(
        self,
        chunk: ParagraphChunk,
    ) -> None:
        """Convert chunk to chunk include chunk."""
        assert chunk.chunk_parent_id and chunk.chunk_parent_name, (
            "Chunk parent ID and name are required (chunk_include_chunk)"
        )

        edge = Edge(
            sid=chunk.chunk_parent_id,
            tid=chunk.chunk_id,
            name=GraphElemType.INCLUDE.value,
            edge_type=GraphElemType.CHUNK_INCLUDE_CHUNK.value,
        )

        self.upsert_edge(
            edges=iter([edge]),
            edge_type=GraphElemType.INCLUDE.value,
            src_type=GraphElemType.CHUNK.value,
            dst_type=GraphElemType.CHUNK.value,
        )

    def upsert_chunk_next_chunk(
        self, chunk: ParagraphChunk, next_chunk: ParagraphChunk
    ):
        """Uperst the vertices and the edge in chunk_next_chunk."""
        edge = Edge(
            sid=chunk.chunk_id,
            tid=next_chunk.chunk_id,
            name=GraphElemType.NEXT.value,
            edge_type=GraphElemType.CHUNK_NEXT_CHUNK.value,
        )

        self.upsert_edge(
            edges=iter([edge]),
            edge_type=GraphElemType.NEXT.value,
            src_type=GraphElemType.CHUNK.value,
            dst_type=GraphElemType.CHUNK.value,
        )

    def upsert_chunk_include_entity(
        self, chunk: ParagraphChunk, entity: Vertex
    ) -> None:
        """Convert chunk to chunk include entity."""
        edge = Edge(
            sid=chunk.chunk_id,
            tid=entity.vid,
            name=GraphElemType.INCLUDE.value,
            edge_type=GraphElemType.CHUNK_INCLUDE_ENTITY.value,
        )

        self.upsert_edge(
            edges=iter([edge]),
            edge_type=GraphElemType.INCLUDE.value,
            src_type=GraphElemType.CHUNK.value,
            dst_type=GraphElemType.ENTITY.value,
        )



"""Text2GQL class."""

import json
import logging
import re
from typing import Dict, List, Union

from dbgpt.core import BaseMessage, HumanPromptTemplate, LLMClient
from dbgpt.rag.transformer.llm_translator import LLMTranslator

TEXT_TO_GQL_PT = """
A question written in graph query language style is provided below. The category of this question, entities and relations that might be used in the cypher query are also provided.
Given the question, translate the question into a cypher query that can be executed on the given knowledge graph. Make sure the syntax of the translated cypher query is correct.
To help query generation, the schema of the knowledge graph is:
{schema}
---------------------
Example:
Question: Query the entity named TuGraph then return the entity.
Category: Single Entity Search
entities: ["TuGraph"]
relations: []
Query:
Match (n) WHERE n.id="TuGraph" RETURN n
Question: Query all one hop paths between the entity named Alex and the entity named TuGraph, then return them.
Category: One Hop Entity Search
entities: ["Alex", "TuGraph"]
relations: []
Query:
MATCH p=(n)-[r]-(m) WHERE n.id="Alex" AND m.id="TuGraph" RETURN p
Question: Query all entities that have a two hop path between them and the entity named Bob, both entities should have a work for relation with the middle entity.
Category: Two Hop Entity Search
entities: ["Bob"]
relations: ["work for"]
Query:
MATCH p=(n)-[r1]-(m)-[r2]-(l) WHERE n.id="Bob" AND r1.id="work for" AND r2.id="work for" RETURN p
Question: Introduce TuGraph and DBGPT seperately.
Category: Freestyle Question
relations: []
Query:
MATCH p=(n)-[r:relation*2]-(m) WHERE n.id IN ["TuGraph", "DB-GPT"] RETURN p
---------------------
Question: {question}
Category: {category}
entities: {entities}
relations: {relations}
Query:

"""  # noqa: E501

logger = logging.getLogger(__name__)


class Text2GQL(LLMTranslator):
    """Text2GQL class."""

    def __init__(self, llm_client: LLMClient, model_name: str):
        """Initialize the Text2GQL."""
        super().__init__(llm_client, model_name, TEXT_TO_GQL_PT)

    def _format_messages(self, text: str, history: str = None) -> List[BaseMessage]:
        # translate intention to gql with single prompt only.
        intention: Dict[str, Union[str, List[str]]] = json.loads(text)
        question = intention.get("rewritten_question", "")
        category = intention.get("category", "")
        entities = intention.get("entities", "")
        relations = intention.get("relations", "")
        schema = intention.get("schema", "")

        template = HumanPromptTemplate.from_template(self._prompt_template)

        messages = (
            template.format_messages(
                schema=schema,
                question=question,
                category=category,
                entities=entities,
                relations=relations,
                history=history,
            )
            if history is not None
            else template.format_messages(
                schema=schema,
                question=question,
                category=category,
                entities=entities,
                relations=relations,
            )
        )

        return messages

    def _parse_response(self, text: str) -> Dict:
        """Parse llm response."""
        translation: Dict[str, str] = {}
        query = ""

        code_block_pattern = re.compile(r"```cypher(.*?)```", re.S)

        result = re.findall(code_block_pattern, text)
        if result:
            query = result[0]
        else:
            query = text

        translation["query"] = query.strip()

        return translation



"""GraphStoreAdapter factory."""

import logging

from dbgpt.storage.graph_store.base import GraphStoreBase
from dbgpt_ext.storage.graph_store.neo4j_store import Neo4jStore
from dbgpt_ext.storage.graph_store.tugraph_store import TuGraphStore
from dbgpt_ext.storage.knowledge_graph.community.base import GraphStoreAdapter
from dbgpt_ext.storage.knowledge_graph.community.neo4j_store_adapter import (
    Neo4jStoreAdapter,
)
from dbgpt_ext.storage.knowledge_graph.community.tugraph_store_adapter import (
    TuGraphStoreAdapter,
)

logger = logging.getLogger(__name__)


class GraphStoreAdapterFactory:
    """Factory for community store adapter."""

    @staticmethod
    def create(graph_store: GraphStoreBase) -> GraphStoreAdapter:
        """Create a GraphStoreAdapter instance.

        Args:
            - graph_store_type: graph store type Memory, TuGraph, Neo4j
        """
        if isinstance(graph_store, TuGraphStore):
            return TuGraphStoreAdapter(graph_store)
        elif isinstance(graph_store, Neo4jStore):
            # Neo4j has its own adapter due to different syntax and data model
            return Neo4jStoreAdapter(graph_store)
        else:
            raise Exception(
                "create community store adapter for %s failed",
                graph_store.__class__.__name__,
            )



"""DuckDB connector."""

from dataclasses import dataclass, field
from typing import Any, Iterable, Optional, Type

from sqlalchemy import create_engine, text

from dbgpt.core.awel.flow import (
    TAGS_ORDER_HIGH,
    ResourceCategory,
    auto_register_resource,
)
from dbgpt.datasource.parameter import BaseDatasourceParameters
from dbgpt.datasource.rdbms.base import RDBMSConnector
from dbgpt.util.i18n_utils import _


@auto_register_resource(
    label=_("DuckDB datasource"),
    category=ResourceCategory.DATABASE,
    tags={"order": TAGS_ORDER_HIGH},
    description=_("In-memory analytical database with efficient query processing."),
)
@dataclass
class DuckDbConnectorParameters(BaseDatasourceParameters):
    """DuckDB connection parameters."""

    __type__ = "duckdb"
    path: str = field(metadata={"help": _("Path to the DuckDB file.")})
    driver: str = field(
        default="duckdb",
        metadata={
            "help": _("Driver name for DuckDB, default is duckdb."),
        },
    )

    def create_connector(self) -> "DuckDbConnector":
        """Create DuckDB connector."""
        return DuckDbConnector.from_parameters(self)

    def db_url(self, ssl=False, charset=None):
        """Get the database URL."""
        return f"{self.driver}:///{self.path}"


class DuckDbConnector(RDBMSConnector):
    """DuckDB connector."""

    db_type: str = "duckdb"
    db_dialect: str = "duckdb"

    @classmethod
    def param_class(cls) -> Type[DuckDbConnectorParameters]:
        """Return the parameter class."""
        return DuckDbConnectorParameters

    @classmethod
    def from_parameters(
        cls, parameters: DuckDbConnectorParameters
    ) -> "DuckDbConnector":
        """Create a new DuckDBConnector from parameters."""
        return cls.from_uri(parameters.path)

    @classmethod
    def from_file_path(
        cls, file_path: str, engine_args: Optional[dict] = None, **kwargs: Any
    ) -> RDBMSConnector:
        """Construct a SQLAlchemy engine from URI."""
        _engine_args = engine_args or {}
        return cls(create_engine("duckdb:///" + file_path, **_engine_args), **kwargs)

    def get_users(self):
        """Get users."""
        with self.session_scope() as session:
            cursor = session.execute(
                text(
                    "SELECT * FROM sqlite_master WHERE type = 'table' AND "
                    "name = 'duckdb_sys_users';"
                )
            )
            users = cursor.fetchall()
            return [(user[0], user[1]) for user in users]

    def get_grants(self):
        """Get grants."""
        return []

    def get_collation(self):
        """Get collation."""
        return "UTF-8"

    def get_charset(self):
        """Get character_set of current database."""
        return "UTF-8"

    def get_table_comments(self, db_name: str):
        """Get table comments."""
        with self.session_scope() as session:
            cursor = session.execute(
                text(
                    """
                    SELECT name, sql FROM sqlite_master WHERE type='table'
                    """
                )
            )
            table_comments = cursor.fetchall()
            return [
                (table_comment[0], table_comment[1]) for table_comment in table_comments
            ]

    def table_simple_info(self) -> Iterable[str]:
        """Get table simple info."""
        _tables_sql = """
                SELECT name FROM sqlite_master WHERE type='table'
            """
        with self.session_scope() as session:
            cursor = session.execute(text(_tables_sql))
            tables_results = cursor.fetchall()
            results = []
            for row in tables_results:
                table_name = row[0]
                _sql = f"""
                    PRAGMA  table_info({table_name})
                """
                cursor_colums = session.execute(text(_sql))
                colum_results = cursor_colums.fetchall()
                table_colums = []
                for row_col in colum_results:
                    field_info = list(row_col)
                    table_colums.append(field_info[1])

                results.append(f"{table_name}({','.join(table_colums)});")
            return results



"""Elasticsearch document store."""

import json
import os
from concurrent.futures import Executor, ThreadPoolExecutor
from typing import List, Optional

from dbgpt.core import Chunk
from dbgpt.storage.base import IndexStoreConfig, logger
from dbgpt.storage.full_text.base import FullTextStoreBase
from dbgpt.storage.vector_store.filters import (
    FilterCondition,
    FilterOperator,
    MetadataFilters,
)
from dbgpt.util import string_utils
from dbgpt.util.executor_utils import blocking_func_to_async
from dbgpt_ext.storage.vector_store.elastic_store import ElasticsearchStoreConfig


class ElasticDocumentStore(FullTextStoreBase):
    """Elasticsearch index store."""

    def __init__(
        self,
        es_config: ElasticsearchStoreConfig,
        name: Optional[str] = "dbgpt",
        k1: Optional[float] = 2.0,
        b: Optional[float] = 0.75,
        executor: Optional[Executor] = None,
    ):
        """Init elasticsearch index store.

        refer https://www.elastic.co/guide/en/elasticsearch/reference/8.9/index-
        modules-similarity.html
        TF/IDF based similarity that has built-in tf normalization and is supposed to
        work better for short fields (like names). See Okapi_BM25 for more details.
        This similarity has the following options:
        """
        super().__init__()
        self._es_config = es_config
        from elasticsearch import Elasticsearch

        self._es_config = es_config
        self._es_url = es_config.uri or os.getenv("ELASTICSEARCH_URL", "localhost")
        self._es_port = es_config.port or os.getenv("ELASTICSEARCH_PORT", "9200")
        self._es_username = es_config.user or os.getenv("ELASTICSEARCH_USER", "elastic")
        self._es_password = es_config.password or os.getenv(
            "ELASTICSEARCH_PASSWORD", "dbgpt"
        )
        self._index_name = name.lower()
        if string_utils.contains_chinese(name):
            bytes_str = name.encode("utf-8")
            hex_str = bytes_str.hex()
            self._index_name = "dbgpt_" + hex_str
        # k1 (Optional[float]): Controls non-linear term frequency normalization
        #             (saturation). The default value is 2.0.
        self._k1 = k1 or 2.0
        # b (Optional[float]): Controls to what degree document length normalizes
        #             tf values. The default value is 0.75.
        self._b = b or 0.75
        if self._es_username and self._es_password:
            self._es_client = Elasticsearch(
                hosts=[f"http://{self._es_url}:{self._es_port}"],
                basic_auth=(self._es_username, self._es_password),
            )
        else:
            self._es_client = Elasticsearch(
                hosts=[f"http://{self._es_url}:{self._es_port}"],
            )
        self._es_index_settings = {
            "analysis": {"analyzer": {"default": {"type": "standard"}}},
            "similarity": {
                "custom_bm25": {
                    "type": "BM25",
                    "k1": self._k1,
                    "b": self._b,
                }
            },
        }
        self._es_mappings = {
            "properties": {
                "content": {
                    "type": "text",
                    "similarity": "custom_bm25",
                },
                "metadata": {
                    # Use object so metadata fields stay queryable for filters
                    "type": "object",
                    "dynamic": True,
                },
            }
        }

        if not self._es_client.indices.exists(index=self._index_name):
            self._es_client.indices.create(
                index=self._index_name,
                mappings=self._es_mappings,
                settings=self._es_index_settings,
            )
        self._executor = executor or ThreadPoolExecutor()

    def is_support_full_text_search(self) -> bool:
        # 重写，避免继承父类的默认实现
        """Support full text search.

           Elasticsearch supports full text search.

        Return:
            bool: True if full text search is supported.
        """
        return True  # Elasticsearch 支持全文检索

    def full_text_search(
        self, text: str, topk: int, filters: Optional[MetadataFilters] = None
    ) -> List[Chunk]:
        # 重写，使用现有的 similar_search_with_scores 方法实现全文检索
        """Full text search in Elasticsearch.

        Args:
            text (str): The query text.
            topk (int): Number of results to return. Default is 10.

        Returns:
            List[Chunk]: Search results as chunks.
        """
        score_threshold = 0.0
        return self.similar_search_with_scores(
            text=text, top_k=topk, score_threshold=score_threshold, filters=filters
        )

    def get_config(self) -> IndexStoreConfig:
        """Get the es store config."""
        return self._es_config

    def load_document(self, chunks: List[Chunk]) -> List[str]:
        """Load document in elasticsearch.

        Args:
            chunks(List[Chunk]): document chunks.

        Return:
            List[str]: chunk ids.
        """
        try:
            from elasticsearch.helpers import bulk
        except ImportError:
            raise ValueError("Please install package `pip install elasticsearch`.")
        es_requests = []
        ids = []
        contents = [chunk.content for chunk in chunks]
        metadatas = [self._normalize_metadata(chunk.metadata) for chunk in chunks]
        chunk_ids = [chunk.chunk_id for chunk in chunks]
        for i, content in enumerate(contents):
            es_request = {
                "_op_type": "index",
                "_index": self._index_name,
                "content": content,
                "metadata": metadatas[i],
                "_id": chunk_ids[i],
            }
            ids.append(chunk_ids[i])
            es_requests.append(es_request)
        bulk(self._es_client, es_requests)
        self._es_client.indices.refresh(index=self._index_name)
        return ids

    def similar_search(
        self, text: str, topk: int, filters: Optional[MetadataFilters] = None
    ) -> List[Chunk]:
        """Search similar text.

        Args:
            text(str): text.
            topk(int): topk.
            filters(MetadataFilters): filters.

        Return:
            List[Chunk]: similar text.
        """
        es_query = self._build_query(text, filters)
        res = self._es_client.search(
            index=self._index_name, body=es_query, size=topk, track_total_hits=False
        )

        chunks = []
        for r in res["hits"]["hits"]:
            metadata = self._normalize_metadata(r["_source"].get("metadata"))
            chunks.append(
                Chunk(
                    chunk_id=r["_id"],
                    content=r["_source"]["content"],
                    metadata=metadata,
                )
            )
        return chunks

    def similar_search_with_scores(
        self,
        text,
        top_k: int = 10,
        score_threshold: float = 0.3,
        filters: Optional[MetadataFilters] = None,
    ) -> List[Chunk]:
        """Search similar text with scores.

        Args:
            text(str): text.
            top_k(int): top k.
            min_score(float): min score.
            filters(MetadataFilters): filters.

        Return:
            List[Tuple[str, float]]: similar text with scores.
        """
        es_query = self._build_query(text, filters)
        res = self._es_client.search(
            index=self._index_name, body=es_query, size=top_k, track_total_hits=False
        )

        chunks_with_scores = []
        for r in res["hits"]["hits"]:
            if r["_score"] >= score_threshold:
                metadata = self._normalize_metadata(r["_source"].get("metadata"))
                chunks_with_scores.append(
                    Chunk(
                        chunk_id=r["_id"],
                        content=r["_source"]["content"],
                        metadata=metadata,
                        score=r["_score"],
                    )
                )
        if score_threshold is not None and len(chunks_with_scores) == 0:
            logger.warning(
                "No relevant docs were retrieved using the relevance score"
                f" threshold {score_threshold}"
            )
        return chunks_with_scores[:top_k]

    async def aload_document(
        self, chunks: List[Chunk], file_id: Optional[str] = None
    ) -> List[str]:
        """Async load document in elasticsearch.

        Args:
            chunks(List[Chunk]): document chunks.
        Return:
            List[str]: chunk ids.
        """
        # 新增修改：将 file_id 注入到每个 chunk 的元数据中
        if file_id:
            # 确保 metadata 字段存在，然后添加或更新 file_id
            for chunk in chunks:
                if not hasattr(chunk, "metadata"):
                    chunk.metadata = {}
                chunk.metadata["file_id"] = file_id
        return await blocking_func_to_async(self._executor, self.load_document, chunks)

    def delete_by_ids(self, ids: str) -> List[str]:
        """Delete document by ids.

        Args:
            ids(List[str]): document ids.
        Return:
            return ids.
        """
        id_list = ids.split(",")
        bulk_body = [
            {"delete": {"_index": self._index_name, "_id": doc_id}}
            for doc_id in id_list
        ]
        self._es_client.bulk(body=bulk_body)
        return id_list

    def delete_vector_name(self, index_name: str):
        """Delete index by name.

        Args:
            index_name(str): The name of index to delete.
        """
        self._es_client.indices.delete(index=self._index_name)

    def _build_query(self, text: str, filters: Optional[MetadataFilters]):
        must_clauses = [{"match": {"content": text}}]
        filter_clause = self._build_metadata_filter(filters)
        if filter_clause:
            must_clauses.append(filter_clause)
        return {"query": {"bool": {"must": must_clauses}}}

    def _build_metadata_filter(self, filters: Optional[MetadataFilters]):
        """Translate MetadataFilters to elasticsearch bool clause."""
        if not filters or not filters.filters:
            return None

        clauses = []
        for f in filters.filters:
            field_name = f"metadata.{f.key}"
            if f.operator == FilterOperator.EQ:
                clauses.append({"term": {field_name: f.value}})
            elif f.operator == FilterOperator.IN:
                values = f.value if isinstance(f.value, list) else [f.value]
                clauses.append({"terms": {field_name: values}})
            elif f.operator == FilterOperator.NE:
                clauses.append({"bool": {"must_not": {"term": {field_name: f.value}}}})
            else:
                logger.warning(
                    "Unsupported filter operator %s for elastic full text search",
                    f.operator,
                )
        if not clauses:
            return None
        if filters.condition == FilterCondition.OR:
            return {"bool": {"should": clauses, "minimum_should_match": 1}}
        return {"bool": {"must": clauses}}

    def _normalize_metadata(self, metadata):
        """Ensure metadata is stored as a dict for downstream consumers."""
        if metadata is None:
            return {}
        if isinstance(metadata, dict):
            return metadata
        if isinstance(metadata, str):
            try:
                return json.loads(metadata)
            except Exception:
                # Fallback to wrapping the raw string to avoid breaking callers
                return {"value": metadata}
        try:
            return dict(metadata)
        except Exception:
            return {"value": metadata}



"""Module Of Knowledge."""

from typing import Any, Dict

from dbgpt_ext.rag.knowledge.factory import KnowledgeFactory

_MODULE_CACHE: Dict[str, Any] = {}


def __getattr__(name: str):
    # Lazy load
    import importlib

    if name in _MODULE_CACHE:
        return _MODULE_CACHE[name]

    _LIBS = {
        "KnowledgeFactory": "factory",
        "CSVKnowledge": "csv",
        "DatasourceKnowledge": "datasource",
        "DocxKnowledge": "docx",
        "Word97DocKnowledge": "doc",
        "HTMLKnowledge": "html",
        "MarkdownKnowledge": "markdown",
        "PDFKnowledge": "pdf",
        "PPTXKnowledge": "pptx",
        "StringKnowledge": "string",
        "TXTKnowledge": "txt",
        "URLKnowledge": "url",
        "ExcelKnowledge": "xlsx",
    }

    if name in _LIBS:
        module_path = "." + _LIBS[name]
        module = importlib.import_module(module_path, __name__)
        attr = getattr(module, name)
        _MODULE_CACHE[name] = attr
        return attr
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")


__all__ = [
    "KnowledgeFactory",
    "CSVKnowledge",
    "DatasourceKnowledge",
    "DocxKnowledge",
    "Word97DocKnowledge",
    "HTMLKnowledge",
    "MarkdownKnowledge",
    "PDFKnowledge",
    "PPTXKnowledge",
    "StringKnowledge",
    "TXTKnowledge",
    "URLKnowledge",
    "ExcelKnowledge",
]



"""Datasource Knowledge."""

from typing import Any, Dict, List, Optional, Union

from dbgpt.core import Document
from dbgpt.datasource import BaseConnector
from dbgpt.rag.knowledge.base import (
    ChunkStrategy,
    DocumentType,
    Knowledge,
    KnowledgeType,
)

from ..summary.rdbms_db_summary import (
    _DEFAULT_COLUMN_SEPARATOR,
    _DEFAULT_SUMMARY_TEMPLATE,
    _parse_db_summary_with_metadata,
)


class DatasourceKnowledge(Knowledge):
    """Datasource Knowledge."""

    def __init__(
        self,
        connector: BaseConnector,
        summary_template: str = _DEFAULT_SUMMARY_TEMPLATE,
        separator: str = "--table-field-separator--",
        column_separator: str = _DEFAULT_COLUMN_SEPARATOR,
        knowledge_type: Optional[KnowledgeType] = KnowledgeType.DOCUMENT,
        metadata: Optional[Dict[str, Union[str, List[str]]]] = None,
        model_dimension: int = 512,
        **kwargs: Any,
    ) -> None:
        """Create Datasource Knowledge with Knowledge arguments.

        Args:
            connector(BaseConnector): connector
            summary_template(str, optional): summary template
            separator(str, optional): separator used to separate
                table's basic info and fields.
                defaults `-- table-field-separator--`
            knowledge_type(KnowledgeType, optional): knowledge type
            metadata(Dict[str, Union[str, List[str]], optional): metadata
            model_dimension(int, optional): The threshold for splitting field string
        """
        self._separator = separator
        self._column_separator = column_separator
        self._connector = connector
        self._summary_template = summary_template
        self._model_dimension = model_dimension
        super().__init__(knowledge_type=knowledge_type, metadata=metadata, **kwargs)

    def _load(self) -> List[Document]:
        """Load datasource document from data_loader."""
        docs = []
        db_summary_with_metadata = _parse_db_summary_with_metadata(
            self._connector,
            self._summary_template,
            self._separator,
            column_separator=self._column_separator,
            model_dimension=self._model_dimension,
        )
        for summary, table_metadata in db_summary_with_metadata:
            metadata = {"source": "database"}

            if self._metadata:
                metadata.update(self._metadata)  # type: ignore
            table_metadata.update(metadata)
            docs.append(Document(content=summary, metadata=table_metadata))
        return docs

    @classmethod
    def support_chunk_strategy(cls) -> List[ChunkStrategy]:
        """Return support chunk strategy."""
        return [
            ChunkStrategy.CHUNK_BY_SIZE,
            ChunkStrategy.CHUNK_BY_SEPARATOR,
            ChunkStrategy.CHUNK_BY_PAGE,
        ]

    @classmethod
    def type(cls) -> KnowledgeType:
        """Knowledge type of Datasource."""
        return KnowledgeType.DOCUMENT

    @classmethod
    def document_type(cls) -> DocumentType:
        """Return document type."""
        return DocumentType.DATASOURCE

    @classmethod
    def default_chunk_strategy(cls) -> ChunkStrategy:
        """Return default chunk strategy.

        Returns:
            ChunkStrategy: default chunk strategy
        """
        return ChunkStrategy.CHUNK_BY_PAGE



from pathlib import Path

import pytest

from ..doc import Word97DocKnowledge


@pytest.fixture
def mock_file_path():
    file_path = Path(__file__).parent / "data" / "test_mock.doc"
    return file_path.as_posix()


def test_load_from_docx(mock_file_path):
    knowledge = Word97DocKnowledge(file_path=mock_file_path)
    documents = knowledge._load()
    actual = documents[0].content.replace("\r", "\n")
    assert len(documents) == 1
    assert actual == "This is the first paragraph.\n\nThis is the second paragraph.\n"
    assert documents[0].metadata["source"] == mock_file_path



"""TuGraph Connector."""

import json
from dataclasses import dataclass, field
from typing import Dict, Generator, Iterator, List, Type, cast

from dbgpt.core.awel.flow import (
    TAGS_ORDER_HIGH,
    ResourceCategory,
    auto_register_resource,
)
from dbgpt.datasource.base import BaseConnector
from dbgpt.datasource.parameter import BaseDatasourceParameters
from dbgpt.util.i18n_utils import _


@auto_register_resource(
    label=_("TuGraph datasource"),
    category=ResourceCategory.DATABASE,
    tags={"order": TAGS_ORDER_HIGH},
    description=_(
        "TuGraph is a high-performance graph database jointly developed by Ant Group "
        "and Tsinghua University."
    ),
)
@dataclass
class TuGraphParameters(BaseDatasourceParameters):
    """TuGraph connection parameters."""

    __type__ = "tugraph"

    host: str = field(metadata={"help": _("TuGraph server host")})
    user: str = field(metadata={"help": _("TuGraph server user")})
    password: str = field(
        default="${env:DBGPT_DB_PASSWORD}",
        metadata={
            "help": _(
                "Database password, you can write your password directly, of course, "
                "you can also use environment variables, such as "
                "${env:DBGPT_DB_PASSWORD}"
            ),
            "tags": "privacy",
        },
    )
    port: int = field(
        default=7687, metadata={"help": _("TuGraph server port, default 7687")}
    )
    database: str = field(
        default="default", metadata={"help": _("Database name, default 'default'")}
    )

    def create_connector(self) -> "BaseConnector":
        """Create TuGraph connector."""
        return TuGraphConnector.from_parameters(self)

    def db_url(self, ssl=False, charset=None):
        """Get the database URL."""
        raise NotImplementedError("TuGraph does not support db_url")


class TuGraphConnector(BaseConnector):
    """TuGraph connector."""

    db_type: str = "tugraph"
    driver: str = "bolt"
    dialect: str = "tugraph"

    def __init__(self, driver, graph):
        """Initialize the connector with a Neo4j driver."""
        self._driver = driver
        self._schema = None
        self._graph = graph
        self._session = None
        self._is_closed = False

    def create_graph(self, graph_name: str) -> bool:
        """Create a new graph in the database if it doesn't already exist."""
        try:
            with self._driver.session(database="default") as session:
                graph_list = session.run("CALL dbms.graph.listGraphs()").data()
                exists = any(item["graph_name"] == graph_name for item in graph_list)
                if not exists:
                    session.run(
                        f"CALL dbms.graph.createGraph('{graph_name}', '', 2048)"
                    )
        except Exception as e:
            raise Exception(f"Failed to create graph '{graph_name}': {str(e)}") from e

        return not exists

    def is_exist(self, graph_name: str) -> bool:
        """Check a new graph in the database if it doesn't already exist."""
        try:
            with self._driver.session(database="default") as session:
                graph_list = session.run("CALL dbms.graph.listGraphs()").data()
                exists = any(item["graph_name"] == graph_name for item in graph_list)
        except Exception as e:
            raise Exception(
                f"Failed to check graph exist'{graph_name}': {str(e)}"
            ) from e

        return exists

    def delete_graph(self, graph_name: str) -> None:
        """Delete a graph in the database if it exists."""
        with self._driver.session(database="default") as session:
            graph_list = session.run("CALL dbms.graph.listGraphs()").data()
            exists = any(item["graph_name"] == graph_name for item in graph_list)
            if exists:
                session.run(f"Call dbms.graph.deleteGraph('{graph_name}')")

    @classmethod
    def param_class(cls) -> Type[TuGraphParameters]:
        """Return the parameter class."""
        return TuGraphParameters

    @classmethod
    def from_parameters(cls, parameters: TuGraphParameters) -> "TuGraphConnector":
        """Create a new TuGraphConnector from parameters."""
        return cls.from_uri_db(
            parameters.host,
            parameters.port,
            parameters.user,
            parameters.password,
            parameters.database,
        )

    @classmethod
    def from_uri_db(
        cls, host: str, port: int, user: str, pwd: str, db_name: str
    ) -> "TuGraphConnector":
        """Create a new TuGraphConnector from host, port, user, pwd, db_name."""
        try:
            from neo4j import GraphDatabase

            db_url = f"{cls.driver}://{host}:{str(port)}"
            driver = GraphDatabase.driver(db_url, auth=(user, pwd))
            driver.verify_connectivity()
            return cast(TuGraphConnector, cls(driver=driver, graph=db_name))

        except ImportError as err:
            raise ImportError(
                "neo4j package is not installed, please install it with "
                "`pip install neo4j`"
            ) from err

    def get_system_info(self) -> Dict:
        """Get system info from the TuGraph."""
        with self._driver.session(database="default") as session:
            system_info_list = session.run("CALL dbms.system.info()")
            system_info = {}
            for info in system_info_list:
                system_info[info["name"]] = info["value"]
            return system_info

    def get_table_names(self) -> Iterator[str]:
        """Get all table names from the TuGraph by Neo4j driver."""
        with self._driver.session(database=self._graph) as session:
            # Run the query to get vertex labels
            raw_vertex_labels = session.run("CALL db.vertexLabels()").data()
            vertex_labels = [
                table_name["label"] + "_vertex" for table_name in raw_vertex_labels
            ]

            # Run the query to get edge labels
            raw_edge_labels = session.run("CALL db.edgeLabels()").data()
            edge_labels = [
                table_name["label"] + "_edge" for table_name in raw_edge_labels
            ]

            return iter(vertex_labels + edge_labels)

    def get_grants(self):
        """Get grants."""
        return []

    def get_collation(self):
        """Get collation."""
        return "UTF-8"

    def get_charset(self):
        """Get character_set of current database."""
        return "UTF-8"

    def table_simple_info(self):
        """Get table simple info."""
        return []

    def close(self):
        """Close the Neo4j driver."""
        if self._is_closed:
            return
        self._driver.close()
        self._is_closed = True

    def run(self, query: str, fetch: str = "all") -> List:
        """Run query."""
        with self._driver.session(database=self._graph) as session:
            try:
                result = session.run(query)
                return list(result)
            except Exception as e:
                raise Exception(f"Query execution failed: {e}\nQuery: {query}") from e

    def run_stream(self, query: str) -> Generator:
        """Run GQL."""
        with self._driver.session(database=self._graph) as session:
            result = session.run(query)
            yield from result

    def get_columns(self, table_name: str, table_type: str = "vertex") -> List[Dict]:
        """Retrieve the column for a specified vertex or edge table in the graph db.

        This function queries the schema of a given table (vertex or edge) and returns
        detailed information about its columns (properties).

        Args:
            table_name (str): table name (graph name)
            table_type (str): table type (vertex or edge)

        Returns:
            columns: List[Dict], which contains name: str, type: str,
                default_expression: str, is_in_primary_key: bool, comment: str
                eg:[{'name': 'id', 'type': 'int', 'default_expression': '',
                'is_in_primary_key': True, 'comment': 'id'}, ...]
        """
        with self._driver.session(database=self._graph) as session:
            data = []
            result = None
            if table_type == "vertex":
                result = session.run(f"CALL db.getVertexSchema('{table_name}')").data()
            else:
                result = session.run(f"CALL db.getEdgeSchema('{table_name}')").data()
            schema_info = json.loads(result[0]["schema"])
            for prop in schema_info.get("properties", []):
                prop_dict = {
                    "name": prop["name"],
                    "type": prop["type"],
                    "default_expression": "",
                    "is_in_primary_key": bool(
                        "primary" in schema_info
                        and prop["name"] == schema_info["primary"]
                    ),
                    "comment": prop["name"],
                }
                data.append(prop_dict)
            return data

    def get_indexes(self, table_name: str, table_type: str = "vertex") -> List[Dict]:
        """Get table indexes about specified table.

        Args:
            table_name (str): table name
            table_type (str): 'vertex' | 'edge'
        Returns:
            List[Dict]:eg:[{'name': 'idx_key', 'column_names': ['id']}]
        """
        # [{'name':'id','column_names':['id']}]
        with self._driver.session(database=self._graph) as session:
            result = session.run(
                f"CALL db.listLabelIndexes('{table_name}','{table_type}')"
            ).data()
            transformed_data = []
            for item in result:
                new_dict = {"name": item["field"], "column_names": [item["field"]]}
                transformed_data.append(new_dict)
            return transformed_data

    @classmethod
    def is_graph_type(cls) -> bool:
        """Return whether the connector is a graph database connector."""
        return True



"""Embedding retriever operator."""

from functools import reduce
from typing import List, Optional, Union

from dbgpt.core import Chunk
from dbgpt.core.awel.flow import IOField, OperatorCategory, Parameter, ViewMetadata
from dbgpt.core.interface.operators.retriever import RetrieverOperator
from dbgpt.rag.knowledge import Knowledge
from dbgpt.rag.retriever.embedding import EmbeddingRetriever
from dbgpt.rag.retriever.rerank import Ranker
from dbgpt.rag.retriever.rewrite import QueryRewrite
from dbgpt.storage.base import IndexStoreBase
from dbgpt.util.i18n_utils import _
from dbgpt_ext.rag.chunk_manager import ChunkParameters

from ..assembler.embedding import EmbeddingAssembler
from .assembler import AssemblerOperator


class EmbeddingRetrieverOperator(RetrieverOperator[Union[str, List[str]], List[Chunk]]):
    """The Embedding Retriever Operator."""

    metadata = ViewMetadata(
        label=_("Embedding Retriever Operator"),
        name="embedding_retriever_operator",
        description=_("Retrieve candidates from vector store."),
        category=OperatorCategory.RAG,
        parameters=[
            Parameter.build_from(
                _("Storage Index Store"),
                "index_store",
                IndexStoreBase,
                description=_("The vector store connector."),
                alias=["vector_store_connector"],
            ),
            Parameter.build_from(
                _("Top K"),
                "top_k",
                int,
                description=_("The number of candidates."),
            ),
            Parameter.build_from(
                _("Score Threshold"),
                "score_threshold",
                float,
                description=_(
                    "The score threshold, if score of candidate is less than it, it "
                    "will be filtered."
                ),
                optional=True,
                default=0.3,
            ),
            Parameter.build_from(
                _("Query Rewrite"),
                "query_rewrite",
                QueryRewrite,
                description=_("The query rewrite resource."),
                optional=True,
                default=None,
            ),
            Parameter.build_from(
                _("Rerank"),
                "rerank",
                Ranker,
                description=_("The rerank."),
                optional=True,
                default=None,
            ),
        ],
        inputs=[
            IOField.build_from(
                _("Query"),
                "query",
                str,
                description=_("The query to retrieve."),
            )
        ],
        outputs=[
            IOField.build_from(
                _("Candidates"),
                "candidates",
                Chunk,
                description=_("The retrieved candidates."),
                is_list=True,
            )
        ],
    )

    def __init__(
        self,
        index_store: IndexStoreBase,
        top_k: int,
        score_threshold: float = 0.3,
        query_rewrite: Optional[QueryRewrite] = None,
        rerank: Optional[Ranker] = None,
        **kwargs,
    ):
        """Create a new EmbeddingRetrieverOperator."""
        super().__init__(**kwargs)
        self._score_threshold = score_threshold
        self._retriever = EmbeddingRetriever(
            index_store=index_store,
            top_k=top_k,
            query_rewrite=query_rewrite,
            rerank=rerank,
        )

    def retrieve(self, query: Union[str, List[str]]) -> List[Chunk]:
        """Retrieve the candidates."""
        if isinstance(query, str):
            return self._retriever.retrieve_with_scores(query, self._score_threshold)
        elif isinstance(query, list):
            candidates = [
                self._retriever.retrieve_with_scores(q, self._score_threshold)
                for q in query
            ]
            return reduce(lambda x, y: x + y, candidates)


class EmbeddingAssemblerOperator(AssemblerOperator[Knowledge, List[Chunk]]):
    """The Embedding Assembler Operator."""

    metadata = ViewMetadata(
        label=_("Embedding Assembler Operator"),
        name="embedding_assembler_operator",
        description=_("Load knowledge and assemble embedding chunks to vector store."),
        category=OperatorCategory.RAG,
        parameters=[
            Parameter.build_from(
                _("Vector Store Connector"),
                "index_store",
                IndexStoreBase,
                description=_("The vector store connector."),
                alias=["vector_store_connector"],
            ),
            Parameter.build_from(
                _("Chunk Parameters"),
                "chunk_parameters",
                ChunkParameters,
                description=_("The chunk parameters."),
                optional=True,
                default=None,
            ),
        ],
        inputs=[
            IOField.build_from(
                _("Knowledge"),
                "knowledge",
                Knowledge,
                description=_("The knowledge to be loaded."),
            )
        ],
        outputs=[
            IOField.build_from(
                _("Chunks"),
                "chunks",
                List[Chunk],
                description=_(
                    "The assembled chunks, it has been persisted to vector store."
                ),
                is_list=True,
            )
        ],
    )

    def __init__(
        self,
        index_store: IndexStoreBase,
        chunk_parameters: Optional[ChunkParameters] = None,
        **kwargs,
    ):
        """Create a new EmbeddingAssemblerOperator.

        Args:
            index_store (IndexStoreBase): The index storage.
            chunk_parameters (Optional[ChunkParameters], optional): The chunk
                parameters. Defaults to ChunkParameters(chunk_strategy="CHUNK_BY_SIZE").
        """
        if not chunk_parameters:
            chunk_parameters = ChunkParameters(chunk_strategy="CHUNK_BY_SIZE")
        self._chunk_parameters = chunk_parameters
        self._index_store = index_store
        super().__init__(**kwargs)

    def assemble(self, knowledge: Knowledge) -> List[Chunk]:
        """Assemble knowledge for input value."""
        assembler = EmbeddingAssembler.load_from_knowledge(
            knowledge=knowledge,
            chunk_parameters=self._chunk_parameters,
            index_store=self._index_store,
        )
        assembler.persist()
        return assembler.get_chunks()



"""Vector Based Graph Retriever."""

import logging
from typing import List, Tuple

from dbgpt.storage.graph_store.graph import Graph
from dbgpt_ext.rag.retriever.graph_retriever.base import GraphRetrieverBase

logger = logging.getLogger(__name__)


class VectorBasedGraphRetriever(GraphRetrieverBase):
    """Vector Based Graph Retriever class."""

    def __init__(
        self,
        graph_store_adapter,
        triplet_topk,
        similarity_search_topk,
        similarity_search_score_threshold,
    ):
        """Initialize Vector Based Graph Retriever."""
        self._graph_store_adapter = graph_store_adapter
        self._triplet_topk = triplet_topk
        self._similarity_search_topk = similarity_search_topk
        self._similarity_search_score_threshold = similarity_search_score_threshold

    async def retrieve(self, vectors: List[List[float]]) -> Tuple[Graph, None]:
        """Retrieve from triplet graph with vectors."""
        subgraph = self._graph_store_adapter.explore_trigraph(
            subs=vectors,
            topk=self._similarity_search_topk,
            limit=self._triplet_topk,
            score_threshold=self._similarity_search_score_threshold,
        )

        return subgraph



"""OpenSPG class."""

import logging
from dataclasses import dataclass

from dbgpt.storage.knowledge_graph.base import KnowledgeGraphBase, KnowledgeGraphConfig

logger = logging.getLogger(__name__)


@dataclass
class OpenSPGConfig(KnowledgeGraphConfig):
    """OpenSPG config."""

    __type__ = "openspg"


class OpenSPG(KnowledgeGraphBase):
    """OpenSPG class."""

    # todo: add OpenSPG implementation



"""Base Linker."""

from abc import ABC, abstractmethod
from typing import List


class BaseSchemaLinker(ABC):
    """Base Linker."""

    def schema_linking(self, query: str) -> List:
        """Query schema info.

        Args:
            query (str): query text
        Returns:
            List: list of schema
        """
        return self._schema_linking(query)

    def schema_linking_with_vector_db(self, query: str) -> List:
        """Query schema info with vector db.

        Args:
            query (str): query text
        Returns:
            List: list of schema
        """
        return self._schema_linking_with_vector_db(query)

    async def schema_linking_with_llm(self, query: str) -> List:
        """Query schema info with llm.

        Args:
            query(str): query text
        Returns:
        List: list of schema
        """
        return await self._schema_linking_with_llm(query)

    @abstractmethod
    def _schema_linking(self, query: str) -> List:
        """Get DB schema info.

        Args:
            query (str): query text
        Returns:
            List: list of schema
        """

    @abstractmethod
    def _schema_linking_with_vector_db(self, query: str) -> List:
        """Query schema info with vector db.

        Args:
            query (str): query text
        Returns:
            List: list of schema
        """

    @abstractmethod
    async def _schema_linking_with_llm(self, query: str) -> List:
        """Query schema info with llm.

        Args:
            query (str): query text
        Returns:
            List: list of schema
        """



"""Base class for Vertica dialect."""

from __future__ import (
    absolute_import,
    annotations,
    division,
    print_function,
    unicode_literals,
)

import logging
import re
from typing import Any, Optional

from sqlalchemy import sql
from sqlalchemy.engine import default, reflection

logger: logging.Logger = logging.getLogger(__name__)


class VerticaInspector(reflection.Inspector):
    """Reflection inspector for Vertica."""

    dialect: VerticaDialect

    def get_all_columns(self, table, schema: Optional[str] = None, **kw: Any):
        r"""Return all table columns names within a particular schema."""
        return self.dialect.get_all_columns(
            self.bind, table, schema, info_cache=self.info_cache, **kw
        )

    def get_table_comment(self, table_name: str, schema: Optional[str] = None, **kw):
        """Return comment of a table in a schema."""
        return self.dialect.get_table_comment(
            self.bind, table_name, schema, info_cache=self.info_cache, **kw
        )

    def get_view_columns(
        self, view: Optional[str] = None, schema: Optional[str] = None, **kw: Any
    ):
        r"""Return all view columns names within a particular schema."""
        return self.dialect.get_view_columns(
            self.bind, view, schema, info_cache=self.info_cache, **kw
        )

    def get_view_comment(
        self, view: Optional[str] = None, schema: Optional[str] = None, **kw
    ):
        r"""Return view comments within a particular schema."""
        return self.dialect.get_view_comment(
            self.bind, view, schema, info_cache=self.info_cache, **kw
        )


class VerticaDialect(default.DefaultDialect):
    """Vertica dialect."""

    name = "vertica"
    inspector = VerticaInspector

    def __init__(self, json_serializer=None, json_deserializer=None, **kwargs):
        """Init object."""
        default.DefaultDialect.__init__(self, **kwargs)

        self._json_deserializer = json_deserializer
        self._json_serializer = json_serializer

    def initialize(self, connection):
        """Init dialect."""
        super().initialize(connection)

    def _get_default_schema_name(self, connection):
        return connection.scalar(sql.text("SELECT current_schema()"))

    def _get_server_version_info(self, connection):
        v = connection.scalar(sql.text("SELECT version()"))
        m = re.match(r".*Vertica Analytic Database v(\d+)\.(\d+)\.(\d)+.*", v)
        if not m:
            raise AssertionError(
                "Could not determine version from string '%(ver)s'" % {"ver": v}
            )
        return tuple([int(x) for x in m.group(1, 2, 3) if x is not None])

    def create_connect_args(self, url):
        """Create args of connection."""
        opts = url.translate_connect_args(username="user")
        opts.update(url.query)
        return [], opts

    def has_table(self, connection, table_name, schema=None):
        """Check availability of a table."""
        return False

    def has_sequence(self, connection, sequence_name, schema=None):
        """Check availability of a sequence."""
        return False

    def has_type(self, connection, type_name):
        """Check availability of a type."""
        return False

    def get_schema_names(self, connection, **kw):
        """Return names of all schemas."""
        return []

    def get_table_comment(self, connection, table_name, schema=None, **kw):
        """Return comment of a table in a schema."""
        return {"text": table_name}

    def get_table_names(self, connection, schema=None, **kw):
        """Get names of tables in a schema."""
        return []

    def get_temp_table_names(self, connection, schema=None, **kw):
        """Get names of temp tables in a schema."""
        return []

    def get_view_names(self, connection, schema=None, **kw):
        """Get names of views in a schema."""
        return []

    def get_view_definition(self, connection, view_name, schema=None, **kw):
        """Get definition of views in a schema."""
        return view_name

    def get_temp_view_names(self, connection, schema=None, **kw):
        """Get names of temp views in a schema."""
        return []

    def get_unique_constraints(self, connection, table_name, schema=None, **kw):
        """Get unique constrains of a table in a schema."""
        return []

    def get_check_constraints(self, connection, table_name, schema=None, **kw):
        """Get checks of a table in a schema."""
        return []

    def normalize_name(self, name):
        """Normalize name."""
        name = name and name.rstrip()
        if name is None:
            return None
        return name.lower()

    def denormalize_name(self, name):
        """Denormalize name."""
        return name

    def get_foreign_keys(self, connection, table_name, schema=None, **kw):
        """Get poreignn keys of a table in a schema."""
        return []

    def get_indexes(self, connection, table_name, schema, **kw):
        """Get indexes of a table in a schema."""
        return []

    def visit_create_index(self, create):
        """Disable index creation since that's not a thing in Vertica."""
        return None

    def get_pk_constraint(self, connection, table_name, schema=None, **kw):
        """Get primary keye of a table in a schema."""
        return None

    def get_all_columns(self, connection, table, schema=None, **kw):
        """Get all columns of a table in a schema."""
        return []

    def get_columns(self, connection, table_name, schema=None, **kw):
        """Get all columns of a table in a schema."""
        return self.get_all_columns(connection, table_name, schema)

    def get_view_columns(self, connection, view, schema=None, **kw):
        """Get columns of views in a schema."""
        return []

    def get_view_comment(self, connection, view, schema=None, **kw):
        """Get comment of view."""
        return {"text": view}
